#!/usr/bin/env python
from __future__ import print_function
import re
import shutil
import unittest
from ruffus.ruffus_utility import CHECKSUM_FILE_TIMESTAMPS, RUFFUS_HISTORY_FILE, \
    get_default_history_file_name
from ruffus.ruffus_exceptions import RethrownJobError, fatal_error_input_file_does_not_match
from ruffus import pipeline_run, pipeline_printout, suffix, transform, split, merge, \
    dbdict, follows, originate, Pipeline, regex
import ruffus
import sys

"""

    test_regex_error_messages.py

        test product, combine, permute, combine_with_replacement

        Includes code from python.unittest with the following copyright notice:


        Copyright (c) 1999-2003 Steve Purcell
        Copyright (c) 2003-2010 Python Software Foundation
        This module is free software, and you may redistribute it and/or modify
        it under the same terms as Python itself, so long as this copyright message
        and disclaimer are retained in their original form.

        IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
        SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF
        THIS CODE, EVEN IF THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
        DAMAGE.

        THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
        LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
        PARTICULAR PURPOSE.  THE CODE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS,
        AND THERE IS NO OBLIGATION WHATSOEVER TO PROVIDE MAINTENANCE,
        SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

"""

import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0]))
# sub-1s resolution in system?
one_second_per_job = None
parallelism = 2


# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


try:
    from StringIO import StringIO
except:
    from io import StringIO


# ___________________________________________________________________________
#
#   generate_initial_files1
# ___________________________________________________________________________
def generate_initial_files1(out_name):
    with open(out_name, 'w') as outfile:
        pass

# ___________________________________________________________________________
#
#   check_regex_task
# ___________________________________________________________________________


def check_regex_task(infiles, outfile,
                     prefix1,
                     prefix2,
                     extension):
    with open(outfile, "w") as p:
        pass

    if prefix1 != prefix2:
        raise Exception("Expecting %s == %s" % (prefix1, prefix2))


# ___________________________________________________________________________
#
#   check_regex_unmatched_task
# ___________________________________________________________________________
def check_regex_unmatched_task(infiles, outfile,
                               prefix1,
                               prefix2,
                               extension):
    raise Exception("Should blow up first")


# ___________________________________________________________________________
#
#   check_suffix_task
# ___________________________________________________________________________
def check_suffix_task(infile, outfile,
                      basename):
    with open(outfile, "w") as f:
        pass


# ___________________________________________________________________________
#
#   check_suffix_unmatched_task
# ___________________________________________________________________________
def check_suffix_unmatched_task(infiles, outfile, unknown):
    raise Exception("Should blow up first")


# ___________________________________________________________________________
#
#   check_suffix_unmatched_task
# ___________________________________________________________________________
def check_suffix_unmatched_task2(infiles, outfile):
    raise Exception("Should blow up first")


# ___________________________________________________________________________
#
#   check_product_misspelt_capture_error_task
# ___________________________________________________________________________
def check_regex_misspelt_capture_error_task(infiles, outfile,
                                            prefix1,
                                            prefix2,
                                            extension):
    raise Exception("Should blow up first")


# ___________________________________________________________________________
#
#   check_regex_misspelt_capture2_error_task
# ___________________________________________________________________________
def check_regex_misspelt_capture2_error_task(infiles, outfile,
                                             prefix1,
                                             prefix2,
                                             extension):
    raise Exception("Should blow up first")


# ___________________________________________________________________________
#
#   check_regex_out_of_range_regex_reference_error_task
# ___________________________________________________________________________
def check_regex_out_of_range_regex_reference_error_task(infiles, outfile,
                                                        prefix1,
                                                        prefix2,
                                                        extension):
    raise Exception("Should blow up first")


test_pipeline = Pipeline("test")

test_pipeline.originate(task_func=generate_initial_files1,
                        output=[tempdir + "/" + prefix + "_name.tmp1" for prefix in "abcdefghi"])

test_pipeline.transform(task_func=check_regex_task,
                        input=generate_initial_files1,
                        filter=regex("(.*)/(?P<PREFIX>[abcd])(_name)(.tmp1)"),
                        output=r"\1/\g<PREFIX>\3.tmp2",  # output file
                        extras=[r"\2",                # extra: prefix = \2
                                r"\g<PREFIX>",        # extra: prefix = \2
                                r"\4"])               # extra: extension
test_pipeline.transform(task_func=check_regex_unmatched_task,
                        input=generate_initial_files1,
                        filter=regex("(.*)/(?P<PREFIX>[abcd])(_name)(.xxx)"),
                        output=r"\1/\g<PREFIXA>\3.tmp2",  # output file
                        extras=[r"\2",                 # extra: prefix = \2
                                r"\g<PREFIX>",         # extra: prefix = \2
                                r"\4"])                # extra: extension

test_pipeline.transform(task_func=check_suffix_task,
                        input=generate_initial_files1,
                        filter=suffix(".tmp1"),
                        output=r".tmp2",           # output file
                        extras=[r"\1"])            # extra: basename

test_pipeline.transform(task_func=check_suffix_unmatched_task,
                        input=generate_initial_files1,
                        filter=suffix(".tmp1"),
                        output=r".tmp2",           # output file
                        extras=[r"\2"])            # extra: unknown

test_pipeline.transform(task_func=check_suffix_unmatched_task2,
                        input=generate_initial_files1,
                        filter=suffix(".tmp2"),
                        output=r".tmp2")           # output file

test_pipeline.transform(task_func=check_regex_misspelt_capture_error_task,
                        input=generate_initial_files1,
                        filter=regex("(.*)/(?P<PREFIX>[abcd])(_name)(.tmp)"),
                        output=r"\1/\g<PREFIXA>\3.tmp2",  # output file
                        extras=[r"\2",                  # extra: prefix = \2
                                r"\g<PREFIX>",          # extra: prefix = \2
                                r"\4"])                 # extra: extension
test_pipeline.transform(task_func=check_regex_misspelt_capture2_error_task,
                        input=generate_initial_files1,
                        filter=regex("(.*)/(?P<PREFIX>[abcd])(_name)(.tmp)"),
                        output=r"\1/\g<PREFIX>\3.tmp2",  # output file
                        extras=[r"\2",                 # extra: prefix = \2
                                r"\g<PREFIXA>",        # extra: prefix = \2
                                r"\4"])                # extra: extension

test_pipeline.transform(task_func=check_regex_out_of_range_regex_reference_error_task,
                        input=generate_initial_files1,
                        filter=regex("(.*)/(?P<PREFIX>[abcd])(_name)(.tmp)"),
                        output=r"\1/\g<PREFIX>\5.tmp2",  # output file
                        extras=[r"\2",                 # extra: prefix = \2
                                r"\g<PREFIX>",         # extra: prefix = \2
                                r"\4"])                # extra: extension


def cleanup_tmpdir():
    os.system('rm -f %s %s' %
              (os.path.join(tempdir, '*'), RUFFUS_HISTORY_FILE))


class _AssertRaisesContext_27(object):
    """A context manager used to implement TestCase.assertRaises* methods.
    Taken from python unittest2.7
    """

    def __init__(self, expected, test_case, expected_regexp=None):
        self.expected = expected
        self.failureException = test_case.failureException
        self.expected_regexp = expected_regexp

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is None:
            try:
                exc_name = self.expected.__name__
            except AttributeError:
                exc_name = str(self.expected)
            raise self.failureException(
                "{0} not raised".format(exc_name))
        if not issubclass(exc_type, self.expected):
            # let unexpected exceptions pass through
            return False
        self.exception = exc_value  # store for later retrieval
        if self.expected_regexp is None:
            return True

        expected_regexp = self.expected_regexp
        if isinstance(expected_regexp, basestring):
            expected_regexp = re.compile(expected_regexp)
        if not expected_regexp.search(str(exc_value)):
            raise self.failureException('"%s" does not match "%s"' %
                                        (expected_regexp.pattern, str(exc_value)))
        return True


class Test_regex_error_messages(unittest.TestCase):
    def setUp(self):
        try:
            os.mkdir(tempdir)
        except OSError:
            pass
        if sys.hexversion < 0x03000000:
            self.assertRaisesRegex = self.assertRaisesRegexp27

        if sys.hexversion < 0x02070000:
            self.assertIn = self.my_assertIn

    def my_assertIn(self, test_string, full_string):
        self.assertTrue(test_string in full_string)

    #

    def assertRaisesRegexp27(self, expected_exception, expected_regexp,
                             callable_obj=None, *args, **kwargs):
        """Asserts that the message in a raised exception matches a regexp.

        Args:
            expected_exception: Exception class expected to be raised.
            expected_regexp: Regexp (re pattern object or string) expected
                    to be found in error message.
            callable_obj: Function to be called.
            args: Extra args.
            kwargs: Extra kwargs.
        """
        context = _AssertRaisesContext_27(
            expected_exception, self, expected_regexp)
        if callable_obj is None:
            return context
        with context:
            callable_obj(*args, **kwargs)

    # ___________________________________________________________________________
    #
    #   test regex() pipeline_printout and pipeline_run
    # ___________________________________________________________________________

    def test_regex_printout(self):
        cleanup_tmpdir()

        s = StringIO()
        test_pipeline.printout(s, [check_regex_task],
                               verbose=5, wrap_width=10000)
        self.assertTrue(re.search(
            'Missing files.*\[{tempdir}/a_name.tmp1, {tempdir}/a_name.tmp2'.format(tempdir=tempdir), s.getvalue(), re.DOTALL))

    def test_regex_run(self):
        """Run transform(...,regex()...)"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        test_pipeline.run([check_regex_task], verbose=0,
                          multiprocess=parallelism, one_second_per_job=one_second_per_job)

    # ___________________________________________________________________________
    #
    #   test regex() pipeline_printout and pipeline_run
    # ___________________________________________________________________________

    def test_regex_unmatched_printout(self):
        cleanup_tmpdir()
        s = StringIO()
        test_pipeline.printout(
            s, [check_regex_unmatched_task], verbose=5, wrap_width=10000)
        self.assertIn("Warning: Input substitution failed:", s.getvalue())
        self.assertIn(
            "File '{tempdir}/a_name.tmp1' does not match regex".format(tempdir=tempdir), s.getvalue())

    def test_regex_unmatched_run(self):
        """Run transform(...,regex()...)"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        test_pipeline.run([check_regex_unmatched_task], verbose=0,
                          multiprocess=parallelism, one_second_per_job=one_second_per_job)

    # ___________________________________________________________________________
    #
    #   test suffix() pipeline_printout and pipeline_run
    # ___________________________________________________________________________

    def test_suffix_printout(self):
        cleanup_tmpdir()

        s = StringIO()
        test_pipeline.printout(
            s, [check_suffix_task], verbose=5, wrap_width=10000)
        self.assertTrue(re.search(
            'Missing files.*\[{tempdir}/a_name.tmp1, {tempdir}/a_name.tmp2'.format(tempdir=tempdir), s.getvalue(), re.DOTALL))

    def test_suffix_run(self):
        """Run transform(...,suffix()...)"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        test_pipeline.run([check_suffix_task], verbose=0,
                          multiprocess=parallelism, one_second_per_job=one_second_per_job)

    # ___________________________________________________________________________
    #
    #   test suffix() pipeline_printout and pipeline_run
    # ___________________________________________________________________________

    def test_suffix_unmatched(self):
        cleanup_tmpdir()
        s = StringIO()
        self.assertRaisesRegex(fatal_error_input_file_does_not_match,
                               "File '.*?' does not match regex\('.*?'\) and pattern '.*?':\n.*invalid group reference",
                               pipeline_printout,
                               s, [check_suffix_unmatched_task],
                               verbose=3)
        self.assertRaisesRegex(RethrownJobError,
                               "File '.*?' does not match regex\('.*?'\) and pattern '.*?':\n.*invalid group reference",
                               test_pipeline.run,
                               [check_suffix_unmatched_task], verbose=0, multiprocess=parallelism)

    # ___________________________________________________________________________
    #
    #   test suffix() pipeline_printout and pipeline_run
    # ___________________________________________________________________________

    def test_suffix_unmatched_printout2(self):
        cleanup_tmpdir()
        s = StringIO()
        test_pipeline.printout(
            s, [check_suffix_unmatched_task2], verbose=5, wrap_width=10000)
        self.assertIn("Warning: Input substitution failed:", s.getvalue())
        self.assertIn(
            "File '{tempdir}/a_name.tmp1' does not match suffix".format(tempdir=tempdir), s.getvalue())

    def test_suffix_unmatched_run2(self):
        """Run transform(...,suffix()...)"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        test_pipeline.run([check_suffix_unmatched_task2], verbose=0,
                          multiprocess=parallelism, one_second_per_job=one_second_per_job)

    # ___________________________________________________________________________
    #
    #   test regex() errors: func pipeline_printout
    # ___________________________________________________________________________

    def test_regex_misspelt_capture_error(self):
        cleanup_tmpdir()
        s = StringIO()
        self.assertRaisesRegex(fatal_error_input_file_does_not_match,
                               "File '.*?' does not match regex\('.*?'\) and pattern '.*?':\n.*unknown group name",
                               test_pipeline.printout,
                               s, [check_regex_misspelt_capture_error_task],
                               verbose=3)
        self.assertRaisesRegex(RethrownJobError,
                               "File '.*?' does not match regex\('.*?'\) and pattern '.*?':\n.*unknown group name",
                               test_pipeline.run,
                               [check_regex_misspelt_capture_error_task], verbose=0)

    # ___________________________________________________________________________
    #
    #   test regex() errors: func pipeline_printout
    # ___________________________________________________________________________
    def test_regex_misspelt_capture2_error(self):
        cleanup_tmpdir()
        s = StringIO()
        self.assertRaisesRegex(fatal_error_input_file_does_not_match,
                               "File '.*?' does not match regex\('.*?'\) and pattern '.*?':\n.*unknown group name",
                               test_pipeline.printout,
                               s, [check_regex_misspelt_capture2_error_task],
                               verbose=3)
        self.assertRaisesRegex(RethrownJobError,
                               "File '.*?' does not match regex\('.*?'\) and pattern '.*?':\n.*unknown group name",
                               test_pipeline.run,
                               [check_regex_misspelt_capture2_error_task], verbose=0, multiprocess=parallelism)

    # ___________________________________________________________________________
    #
    #   test regex() errors: func pipeline_printout
    # ___________________________________________________________________________

    def test_regex_out_of_range_regex_reference_error_printout(self):
        cleanup_tmpdir()
        s = StringIO()
        self.assertRaisesRegex(fatal_error_input_file_does_not_match,
                               "File '.*?' does not match regex\('.*?'\) and pattern '.*?':\n.*invalid group reference",
                               test_pipeline.printout,
                               s, [check_regex_out_of_range_regex_reference_error_task],
                               verbose=3)
        self.assertRaisesRegex(RethrownJobError,
                               "File '.*?' does not match regex\('.*?'\) and pattern '.*?':\n.*invalid group reference",
                               test_pipeline.run,
                               [check_regex_out_of_range_regex_reference_error_task], verbose=0, multiprocess=parallelism)

    # ___________________________________________________________________________
    #
    #   cleanup
    # ___________________________________________________________________________

    def tearDown(self):
        pass
        shutil.rmtree(tempdir)


#
#   Necessary to protect the "entry point" of the program under windows.
#       see: http://docs.python.org/library/multiprocessing.html#multiprocessing-programming
#
if __name__ == '__main__':
    #pipeline_printout(sys.stdout, [check_product_task], verbose = 3, pipeline= "main")
    parallelism = 1
    suite = unittest.TestLoader().loadTestsFromTestCase(Test_regex_error_messages)
    unittest.TextTestRunner(verbosity=1).run(suite)
    parallelism = 2
    suite = unittest.TestLoader().loadTestsFromTestCase(Test_regex_error_messages)
    unittest.TextTestRunner(verbosity=1).run(suite)
    # unittest.main()



from __future__ import print_function
import copy
import os
from itertools import chain
from . import dbdict
import operator
import marshal
import hashlib
import multiprocessing.managers
from collections import defaultdict
from .ruffus_exceptions import *
from functools import reduce
import glob
import types
import sys
import re
if sys.hexversion < 0x03000000:
    from future_builtins import zip
################################################################################
#
#   ruffus_utility.py
#
#
#   Copyright (c) 10/9/2009 Leo Goodstadt
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#################################################################################


"""

********************************************
:mod:`ruffus_utility` -- Overview
********************************************


.. moduleauthor:: Leo Goodstadt <ruffus@llew.org.uk>

    Common utility functions


"""


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#import task
try:
    from collections.abc import Callable
except ImportError:
    from collections import Callable
try:
    import cPickle as pickle
except:
    import pickle as pickle
if sys.hexversion >= 0x03000000:
    # everything is unicode in python3
    path_str_type = str
else:
    path_str_type = basestring

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Constants


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#
# file to store history out to
#
RUFFUS_HISTORY_FILE = '.ruffus_history.sqlite'
# If DEFAULT_RUFFUS_HISTORY_FILE is specified in the environment variables, use that instead
if "DEFAULT_RUFFUS_HISTORY_FILE" in os.environ:
    RUFFUS_HISTORY_FILE = os.environ["DEFAULT_RUFFUS_HISTORY_FILE"]


# only rerun when the file timestamps are out of date (classic mode)
CHECKSUM_FILE_TIMESTAMPS = 0
# also rerun when the history shows a job as being out of date
CHECKSUM_HISTORY_TIMESTAMPS = 1
CHECKSUM_FUNCTIONS = 2     # also rerun when function body has changed
# also rerun when function parameters or function body change
CHECKSUM_FUNCTIONS_AND_PARAMS = 3

CHECKSUM_REGENERATE = 2     # regenerate checksums

# _________________________________________________________________________________________

#   t_extra_inputs
#       namespaced enum

# _________________________________________________________________________________________


class t_extra_inputs:
    (ADD_TO_INPUTS, REPLACE_INPUTS, KEEP_INPUTS, KEEP_OUTPUTS) = list(range(4))


class inputs(object):
    def __init__(self, *args):
        self.args = args

    def __repr__(self, *args):
        return 'inputs%r' % (self.args,)


class add_inputs(object):
    def __init__(self, *args):
        self.args = args

    def __repr__(self, *args):
        return 'add_inputs%r' % (self.args,)


def get_default_checksum_level():
    """
    Use the checksum level from the environmental variable DEFAULT_RUFFUS_CHECKSUM_LEVEL
    Otherwise default to CHECKSUM_HISTORY_TIMESTAMPS
    """

    #
    #   environmental variable not set
    #
    if "DEFAULT_RUFFUS_CHECKSUM_LEVEL" not in os.environ:
        return CHECKSUM_HISTORY_TIMESTAMPS

    #
    # lookup value from list of CHECKSUM_XXX constants
    #
    checksum_level = None
    env_checksum_level = os.environ["DEFAULT_RUFFUS_CHECKSUM_LEVEL"]
    if len(env_checksum_level) == 1 and env_checksum_level in "0123":
        checksum_level = int(env_checksum_level)
    else:
        global_var = globals()
        for key in global_var:
            if key.startswith('CHECKSUM') and global_var[key] == env_checksum_level:
                checksum_level = value

    #
    #   check environmental variable is valid string
    #
    if checksum_level is None:
        raise error_checksum_level(("The environmental value "
                                    "DEFAULT_RUFFUS_CHECKSUM_LEVEL should be: [0-3 | "
                                    "CHECKSUM_FILE_TIMESTAMPS | "
                                    "CHECKSUM_HISTORY_TIMESTAMPS | "
                                    "CHECKSUM_FUNCTIONS | "
                                    "CHECKSUM_FUNCTIONS_AND_PARAMS] (rather than '%s') ")
                                   % (env_checksum_level,))

    return checksum_level


# _________________________________________________________________________________________

#   open_job_history

# _________________________________________________________________________________________
def get_default_history_file_name():
    history_file = RUFFUS_HISTORY_FILE
    #
    #   try path expansion using the main script name
    #
    try:
        import __main__ as main
        path_parts = path_decomposition(os.path.abspath(main.__file__))
        history_file = history_file.format(**path_parts)
    except Exception:
        pass
    return history_file


def open_job_history(history_file):
    """
    Given a history file name, opens the correspond sqllite db file and returns the handle
    """
    if not history_file:
        history_file = get_default_history_file_name()

    return dbdict.open(history_file, picklevalues=True)


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Functions


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

class JobHistoryChecksum:
    """Class to remember exactly how an output file was created and when."""

    def __str__(self):
        from time import strftime, gmtime
        if hasattr(self, "params"):
            return str([self.outfile,
                        strftime("%d %b %Y %H:%M:%S", gmtime(self.mtime)),
                        self.params,
                        self.task_name
                        ])
        else:
            return strftime("%d %b %Y %H:%M:%S", gmtime(self.mtime))

    def __init__(self, outfile, mtime, params, task):
        # filename and modification time
        self.outfile = outfile
        self.mtime = mtime

        # Uncomment next two lines to debug:
        #self.params = params
        #self.task_name = task._name

        # checksum exact params used to generate this output file
        self.chksum_params = hashlib.md5(pickle.dumps(params)).hexdigest()
        # checksum the function bytecode as well as the function context
        # Don't use func_code alone-- changing the line number of the function,
        # what global variables are available, etc would all change the checksum
        if sys.hexversion >= 0x03000000:
            code = task.user_defined_work_func.__code__
            func_defaults = task.user_defined_work_func.__defaults__
        else:
            code = task.user_defined_work_func.func_code
            func_defaults = task.user_defined_work_func.func_defaults
        func_code = marshal.dumps(code.co_code)

        #
        #   pickle code very defensively, but hopefully without breaking Jake Biesinger's pipelines!
        #
        attributes_to_pickle = [func_defaults,
                                code.co_argcount,
                                code.co_consts,
                                code.co_names,
                                code.co_nlocals,
                                code.co_varnames]

        pickle_results = []
        for aa in attributes_to_pickle:
            # Can't cpickle nested functions: typically blows up with func_code.co_consts
            try:
                pickle_results.append(pickle.dumps(aa))
                continue
            except:
                pass
            # Marshal seems to be less sensitive: try that
            try:
                pickle_results.append(marshal.dumps(aa))
                continue
            except:
                pass
            # Just make a string out of the attribute
            try:
                pickle_results.append(str(aa))
                continue
            except:
                pass
            # OK give up, do nothing: On your head it is

        func_extras = reduce(operator.add, pickle_results)
        self.chksum_func = hashlib.md5(func_code + func_extras).hexdigest()


# _________________________________________________________________________________________
#
#   parameter_list_as_string
#
# _________________________________________________________________________________________
def parameter_list_as_string(parameters):
    """
    Input list of parameters
       Turn this into a string for display

        E.g.

    """
    if parameters is None:
        return ""
    elif not isinstance(parameters, list):
        raise Exception("Unexpected parameter list %s" % (parameters,))
    else:
        return str(parameters)[1:-1]

# _________________________________________________________________________________________
#
#   path_decomposition
#
# _________________________________________________________________________________________


def path_decomposition(orig_path):
    """
    returns a dictionary identifying the components of a file path:
        This has the following keys
            basename: (any) base (file) name of the path not including the extension. No slash included
            ext:      (any) extension of the path including the "."
            path:     a list of subpaths created by removing subdirectory names
            subdir:   a list of subdirectory names from the most nested to the root
        For example
            apath = "/a/b/c/d/filename.txt"
            {   'basename': 'filename',
                'ext':      '.txt'
                'path':     ['/a/b/c/d', '/a/b/c', '/a/b', '/a', '/'], ,
                'subdir': ['d', 'c', 'b', 'a', '/']
            }
            "{path[2]}/changed/{subdir[0]}".format(**res) = '/a/b/changed/d'
            "{path[3]}/changed/{subdir[1]}".format(**res) = '/a/changed/c'
    """
    def recursive_split(a_path):
        """
        split the path into its subdirectories recursively
        """
        if not len(a_path):
            return [[], []]
        if a_path == "/" or a_path == "//":
            return [[a_path], [a_path]]
        sub_path_part, sub_dir_part = os.path.split(a_path)
        if sub_dir_part:
            sub_path_parts, sub_dir_parts = recursive_split(sub_path_part)
            return [[a_path] + sub_path_parts,
                    [sub_dir_part] + sub_dir_parts]
        else:
            return [[], ["/"]]
    #
    if not len(orig_path):
        return {'path': [], 'basename': '', 'ext': '', 'subdir': []}

    # stop normpath from being too clever and removing initial ./ and terminal slash, turning paths into filenames
    if orig_path in ["./", "/."]:
        a_path = orig_path
    else:
        a_path = os.path.normpath(orig_path)
        if orig_path[0:2] == "./" and a_path[0:2] != "./":
            a_path = "./" + a_path

        if orig_path[-1] == "/" and a_path[-1:] != "/":
            a_path += "/"

    path_part, file_part = os.path.split(a_path)
    file_part, ext_part = os.path.splitext(file_part)
    subpaths, subdirs = recursive_split(path_part)
    return {'basename': file_part,
            'ext':      ext_part,
            'subpath':  subpaths,
            'subdir':   subdirs,
            'path':     path_part}


# _________________________________________________________________________________________
#
#   get_nth_nested_level_of_path
#
# _________________________________________________________________________________________
def get_nth_nested_level_of_path(orig_path, n_levels):
    """
    Return path with up to N levels of subdirectories
    0 = full path
    N = 1 : basename
    N = 2 : basename + one subdirectory

    For example
        0   /test/this/now/or/not.txt
        1   not.txt
        2   or/not.txt
        3   now/or/not.txt
        4   this/now/or/not.txt
        5   test/this/now/or/not.txt
        6   /test/this/now/or/not.txt
        7   /test/this/now/or/not.txt
    """
    # FIXME: consider returning full path to make debugging easier or at least
    # make it optional
    if not n_levels or n_levels < 0:
        return orig_path
    res = path_decomposition(orig_path)
    basename = res["basename"] + res["ext"]
    shortened_path = os.path.join(
        *(list(reversed(res["subdir"][0:(n_levels - 1)]))+[basename]))
    if len(shortened_path) < len(orig_path):
        return ".../" + shortened_path


# _________________________________________________________________________________________
#
#   swap_nesting_order
#
# _________________________________________________________________________________________
def swap_nesting_order(orig_coll):
    """
    Reverse nested order so that coll[3]['a'] becomes coll['a'][3]
    """
    new_dict = defaultdict(dict)
    new_list = []
    for ii, ii_item in enumerate(orig_coll):
        for jj, value in ii_item.items():
            if isinstance(jj, int):
                # resize
                new_list += [{}]*(jj + 1 - len(new_list))
                new_list[jj][ii] = value
            else:
                new_dict[jj][ii] = value
    return new_list, dict(new_dict)

# _________________________________________________________________________________________
#
#   swap_doubly_nested_order
#
# _________________________________________________________________________________________


def swap_doubly_nested_order(orig_coll):
    """
    Reverse nested order so that coll[3]['a'] becomes coll['a'][3]
    """
    new_dict = dict()
    new_list = []
    for ii, ii_item in enumerate(orig_coll):
        for jj, jj_item in enumerate(ii_item):
            for kk, value in jj_item.items():
                if isinstance(kk, int):
                    # resize
                    new_list += [{}]*(kk + 1 - len(new_list))
                    if ii not in new_list[kk]:
                        new_list[kk][ii] = dict()
                    new_list[kk][ii][jj] = value
                else:
                    if kk not in new_dict:
                        new_dict[kk] = dict()
                    if ii not in new_dict[kk]:
                        new_dict[kk][ii] = dict()
                    new_dict[kk][ii][jj] = value

    return new_list, new_dict


# _________________________________________________________________________________________
#
#   regex_matches_as_dict
#
# _________________________________________________________________________________________
def regex_matches_as_dict(test_str, compiled_regex):
    """
    Returns result of regular expression match in a dictionary
        combining both named and unnamed captures
    """
    if compiled_regex:
        if isinstance(compiled_regex, path_str_type):
            compiled_regex = re.compile(compiled_regex)
        mm = compiled_regex.search(test_str)
        # Match failed
        if mm is None:
            return False
        else:
            # No capture
            if mm.lastindex is None:
                return {0: mm.group(0)}
            # Combined named and unnamed captures
            else:
                # no dictionary comprehensions in python 2.6 :-(
                #matchdicts.append({i : mm.group(i) for i in (range(mm.lastindex) + mm.groupdict().keys())})
                #   Keys for captures:
                #       1) unnamed captures = range(mm.lastindex + 1)
                #       2) named captures   = mm.groupdict().keys()
                return dict((i, mm.group(i)) for i in (chain(iter(range(mm.lastindex + 1)),
                                                             iter(mm.groupdict().keys()))))

    else:
        return None


# _________________________________________________________________________________________
#
#   path_decomposition_regex_match
#
# _________________________________________________________________________________________
def path_decomposition_regex_match(test_str, compiled_regex):
    """
    Returns a dictionary identifying the components of a file path.

    This includes both the components of a path:
        basename: (any) base (file) name of the path not including the extension. No slash included
        ext:      (any) extension of the path including the "."
        path:     a list of subpaths created by removing subdirectory names
        subdir:   a list of subdirectory names from the most nested to the root
    and regular expression matches
        The keys are the index or name of the capturing group.


    If compiled_regexes is not specified, return path decomposition only

    If compiled_regexes is specified, and the corresponding regular expression does not match,
        the entire match fails

    For example

        path_decomposition_regex_match("/a/b/c/sample1.bam", r"(.*)(?P<id>\d+)\..+")

            {
                0:          '/a/b/c/sample1.bam',           // captured by index
                1:          '/a/b/c/sample',                // captured by index
                'id':       '1'                             // captured by name
                'ext':      '.bam',
                'subdir':   ['c', 'b', 'a', '/'],
                'subpath':  ['/a/b/c', '/a/b', '/a', '/'],
                'path':     '/a/b/c',
                'basename': 'sample1',
            },

        path_decomposition_regex_match("dbsnp15.vcf", r"(.*)(?P<id>\d+)\..+")
            {
                0: 'dbsnp15.vcf',                           // captured by index
                1: 'dbsnp1',                                // captured by index
                'id': '5'                                   // captured by name
                'ext': '.vcf',
                'subdir': [],
                'path': [],
                'basename': 'dbsnp15',
            },


        // fail
        path_decomposition_regex_match("/test.txt", r"(.*)(?P<id>\d+)\..+")
            {}

        // path components only
        path_decomposition_regex_match("/test.txt", None)
            {
                'ext': '.txt',
                'subdir': ['/']
                'subpath': ['/'],
                'path': '/',
                'basename': 'test',
            }

    """
    pp = path_decomposition(test_str)

    # regular expression not specified
    # just path
    if compiled_regex is None:
        return pp

    rr = regex_matches_as_dict(test_str, compiled_regex)

    # regular expression match failed
    # nothing
    if rr == False:
        return {}

    #
    #   regular expression matches override file decomposition values in
    #       case of clashes between predefined keys such as "basename" and
    #       regular expression named capture groups
    #
    pp.update(rr)
    return pp


# _________________________________________________________________________________________
#
#   check_compiled_regexes
#
# _________________________________________________________________________________________
def check_compiled_regexes(compiled_regexes, expected_num):
    """
    check compiled_regexes are of the right type and number
    """
    if compiled_regexes is None:
        return [None] * expected_num

    if not isinstance(compiled_regexes, list):
        raise Exception("Expecting list of None and strings")

    #   pad compiled_regexes with None
    if len(compiled_regexes) < expected_num:
        compiled_regexes.extend(
            [None] * (expected_num - len(compiled_regexes)))

    #   Turn strings to regular expression just in case
    #   We don't want to do this here because the error messages are not very nice:
    #   There is not much context left
    compiled_regexes = [re.compile(rr) if isinstance(
        rr, path_str_type) else rr for rr in compiled_regexes]

    #   check types
    regex_types = type(re.compile("")), type(None)
    for rr in compiled_regexes:
        if not isinstance(rr, regex_types):
            raise Exception(
                "Unexpected type %s ('%s') specified in regular expression list. Expecting string or compiled regular expression" % (type(rr), rr))

    return compiled_regexes


# _________________________________________________________________________________________
#
#   get_all_paths_components
#
# _________________________________________________________________________________________
def get_all_paths_components(paths, compiled_regexes):
    """
        For each path in a list,
            If any of the regular expression matches fails, the whole list fails
    """
    #
    #   merge regular expression matches and path decomposition
    #
    compiled_regexes = check_compiled_regexes(compiled_regexes, len(paths))
    results = []
    for (pp, rr) in zip(paths, compiled_regexes):
        result = path_decomposition_regex_match(pp, rr)
        if result == {}:
            return [{}] * len(paths)
        results.append(result)
    return results


# _________________________________________________________________________________________
#
#   apply_func_to_sequence
#
# _________________________________________________________________________________________
def apply_func_to_sequence(seq, func, tuple_of_conforming_types=(path_str_type,), tuple_of_sequences_types=(list, tuple, set)):
    """
    Recurses into list/tuple/set sequences to apply func to conforming types
    Non-conforming types are left alone
    """
    if isinstance(seq, tuple_of_conforming_types):
        return func(seq)
    elif isinstance(seq, tuple_of_sequences_types):
        return type(seq)(apply_func_to_sequence(pp, func, tuple_of_conforming_types, tuple_of_sequences_types) for pp in seq)
    else:
        return seq


# _________________________________________________________________________________________
#
#   t_regex_replace
#
# _________________________________________________________________________________________
class t_regex_replace(object):
    def __init__(self, filename, regex_str, compiled_regex, regex_or_suffix):
        self.regex_or_suffix = regex_or_suffix
        self.compiled_regex = compiled_regex
        self.regex_str = regex_str
        self.filename = filename

    def __call__(self, p):
        #
        #   check if substitution pattern is mis-specified
        #
        if "\1"in p or "\2" in p:
            raise error_unescaped_regular_expression_forms("['%s'] " % (p.replace("\1", r"\1").replace("\2", r"\2")) +
                                                           "The special regular expression characters "
                                                           r"\1 and \2 need to be 'escaped' in python. "
                                                           r"The easiest option is to use python 'raw' strings "
                                                           r"e.g. r'\1_in_a string\2'. See http://docs.python.org/library/re.html.")
        #
        #   For suffix(), replaces the suffix part by adding leading r"\1" to the substitution pattern
        #
        #   If r"\1" is specified, then we presume you know what you are doing...
        #
        if self.regex_or_suffix == SUFFIX_SUBSTITUTE:
            if r"\1" not in p and r"\g<1>" not in p:
                match_p = r"\g<1>" + p
            else:
                match_p = p

            # throw exception if doesn't match regular expression at all
            (res_str, cnt_replacements) = self.compiled_regex.subn(
                match_p, self.filename)
            if cnt_replacements == 0:
                raise error_input_file_does_not_match(
                    "File '%s' does not match suffix('%s') and pattern '%s'" % (self.filename, self.regex_str, p))
            return res_str

        #
        #   Normal substitution
        #
        #   For suffix(), complete replacement by the specified pattern text
        #           only substitute if r"\1" or r"\g<1>" is specified
        #
        #
        err_str = ""
        try:
            (res_str, cnt_replacements) = self.compiled_regex.subn(p, self.filename)
            if cnt_replacements > 0:
                return res_str
        except re.error:
            exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
            err_str = str(exceptionValue)
            raise fatal_error_input_file_does_not_match(
                "File '%s' does not match regex('%s') and pattern '%s':\n\t%s\n" % (self.filename, self.regex_str, p, err_str))
        except IndexError:
            exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
            err_str = str(exceptionValue)
            raise fatal_error_input_file_does_not_match(
                "File '%s' does not match regex('%s') and pattern '%s':\n\t%s\n" % (self.filename, self.regex_str, p, err_str))

        # except (re.error, IndexError):
            #err_str = str(sys.exc_info()[1]),

        raise error_input_file_does_not_match("File '%s' does not match regex('%s') and pattern '%s'\n%s\n" % (
            self.filename, self.regex_str, p, err_str))


# _________________________________________________________________________________________
#
#   raise_formatter_substitution_exception
#
# _________________________________________________________________________________________
def raise_formatter_substitution_exception(exceptionValue, formatter_str, pattern, filenames,
                                           substitutes_list, substitutes_dict):
    """
    Throws an exception when formatter fails to make a substitution
    """
    # convert to string to get just the missing key
    missing_key = str(exceptionValue.args[0])
    # strip quotes
    if missing_key[0:1] in '\'"' and missing_key[-1:] in '\'"':
        missing_key = missing_key[1:-1]
    raise error_input_file_does_not_match("Unmatched field {%s} in '%s' where\n  input =  %r,\n"
                                          "  filter = formatter(%s). Possible substitutions= %s, %s."
                                          % (missing_key,
                                             pattern,
                                             filenames,
                                             formatter_str,
                                             substitutes_list, substitutes_dict))


# _________________________________________________________________________________________
#
#   t_formatter_replace
#
# _________________________________________________________________________________________
class t_formatter_replace(object):
    def __init__(self, filenames, regex_strings, compiled_regexes=None):
        self.filenames = filenames
        # get the full absolute, normalised paths
        filenames = [os.path.abspath(f) for f in filenames]
        self.path_regex_components = get_all_paths_components(
            filenames, compiled_regexes)
        self.display_regex_strings = parameter_list_as_string(regex_strings)

    def __call__(self, pattern):
        # swapped nesting order makes the syntax easier to explain:
        #   The first level of indirection is always the path component
        #   So basename[0] is the file name for the first file
        #   This looks better than the normal 0[basename]

        # some contortions because format decodes {0} as an offset into a list and not not a lookup into a dict...
        substitutes_list, substitutes_dict = swap_nesting_order(
            self.path_regex_components)

        try:
            return pattern.format(*substitutes_list, **substitutes_dict)
        except (KeyError, IndexError):
            raise_formatter_substitution_exception(sys.exc_info()[1], self.display_regex_strings,
                                                   pattern, self.filenames,
                                                   substitutes_list, substitutes_dict)


# _________________________________________________________________________________________
#
#   t_nested_formatter_replace
#
# _________________________________________________________________________________________
class t_nested_formatter_replace(object):
    """
    Like  t_formatter_replace but with one additional level of nesting
    I.e. everything is a list comprehension!
    For combinatorics @decorators
    """

    def __init__(self, filenames, regex_strings, compiled_regexes):
        # make sure that we have the same level of nestedness for regular expressions and file names etc.
        if len(filenames) != len(regex_strings) or len(filenames) != len(compiled_regexes):
            raise Exception("Logic Error.")
        self.filenames = filenames
        # get the full absolute, normalised paths
        filenames = [[os.path.abspath(f) for f in filegroups]
                     for filegroups in filenames]
        self.path_regex_components = [get_all_paths_components(
            f, r) for (f, r) in zip(filenames, compiled_regexes)]
        self.display_regex_strs = [
            parameter_list_as_string(ss) for ss in regex_strings]

    def __call__(self, pattern):
        # swapped nesting order makes the syntax easier to explain:
        #   The first level of indirection is always the path component
        #   So basename[0] is the file name for the first file
        #   This looks better than the normal 0[basename]

        # some contortions because format decodes {0} as an offset into a list and not not a lookup into a dict...
        substitutes_list, substitutes_dict = swap_doubly_nested_order(
            self.path_regex_components)
        try:
            return pattern.format(*substitutes_list, **substitutes_dict)
        except (KeyError, IndexError):
            formatter_str = ", ".join("formatter(%s)" %
                                      ss for ss in self.display_regex_strs)
            raise_formatter_substitution_exception(sys.exc_info()[1], formatter_str, pattern, self.filenames,
                                                   substitutes_list, substitutes_dict)

# _________________________________________________________________________________________
#
#   t_nested_string_replace
#
# _________________________________________________________________________________________


class t_nested_string_replace(object):
    """
    Replaces path with directory
    """

    def __init__(self, prev_str, new_str):
        self.prev_str = prev_str
        self.new_str = new_str

    def __call__(self, p):
        return p.replace(prev_str, new_str)


# _________________________________________________________________________________________
#
#   regex_replace
#
# _________________________________________________________________________________________

#
#   Perform normal regular expression substitution
#
REGEX_SUBSTITUTE = 0
#
#   An extra peculiar mode to help suffix along:
#   Suffix regular expression have an implicit capture for everything up to the specified
#       suffix text

#
#   By default, replaces the suffix part by adding leading r"\1" to the substitution pattern
#       If r"\1" is already specified in the pattern, then we presume you know what
#       you are doing, and will let you get along with it
#
SUFFIX_SUBSTITUTE = 1

#
#   REGEX_SUBSTITUTE is used for suffix() matches in 'extra' arguments (additional to output)
#   which are strings
#
#   Complete replacement happens. If you wish to retain the prefix text
#       before the suffix, you can do so by adding r"\1"
#


def regex_replace(filename, regex_str, compiled_regex, substitution_patterns, regex_or_suffix=REGEX_SUBSTITUTE):
    return apply_func_to_sequence(substitution_patterns, t_regex_replace(filename, regex_str, compiled_regex, regex_or_suffix))


def formatter_replace(filenames, regex_str, compiled_regex, substitution_patterns):
    return apply_func_to_sequence(substitution_patterns, t_formatter_replace(filenames, regex_str, compiled_regex))


def nested_formatter_replace(filenames, regex_strings, compiled_regexes, substitution_patterns):
    return apply_func_to_sequence(substitution_patterns, t_nested_formatter_replace(filenames, regex_strings, compiled_regexes))


def nested_string_replace(prev_str, new_str, substitution_patterns):
    return apply_func_to_sequence(substitution_patterns, t_nested_string_replace(prev_str, new_str))


# _________________________________________________________________________________________

#   non_str_sequence

# _________________________________________________________________________________________
def non_str_sequence(arg):
    """
    Whether arg is a sequence.
    We treat strings / dicts however as a singleton not as a sequence

    """
    # will only dive into list and set, everything else is not regarded as a sequence
    # loss of flexibility but more conservative
    # if (isinstance(arg, (basestring, dict, multiprocessing.managers.DictProxy))):
    if (not isinstance(arg, (list, tuple, set))):
        return False
    try:
        test = iter(arg)
        return True
    except TypeError:
        return False

# _________________________________________________________________________________________

#   get_strings_in_flattened_sequence_aux

#       helper function for next function

# _________________________________________________________________________________________


def get_strings_in_flattened_sequence_aux(p, l=None):
    """
    Unravels arbitrarily nested sequence and returns lists of strings
    """
    if l is None:
        l = []
    if isinstance(p, path_str_type):
        l.append(p)
    elif non_str_sequence(p):
        for pp in p:
            get_strings_in_flattened_sequence_aux(pp, l)
    return l


# _________________________________________________________________________________________

#   non_str_sequence

# _________________________________________________________________________________________
def get_strings_in_flattened_sequence(p):
    """
    Traverses nested sequence and for each element, returns first string encountered
    """
    if p is None:
        return []

    #
    #  string is returned as list of single string
    #
    if isinstance(p, path_str_type):
        return [p]

    #
    #  Get all strings flattened into list
    #
    return get_strings_in_flattened_sequence_aux(p)


# _________________________________________________________________________________________

#   get_first_string_in_nested_sequence

# _________________________________________________________________________________________
def get_first_string_in_nested_sequence(p):
    strings = get_strings_in_flattened_sequence(p)
    if len(strings):
        return strings[0]
    return None


#
#   TODOOO third object could be a dict or a list
#
# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Encoders: turn objects and filenames into a more presentable format

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
def ignore_unknown_encoder(obj):
    if non_str_sequence(obj):
        return "[%s]" % ", ".join(map(ignore_unknown_encoder, obj))
    try:
        s = str(obj)
        if " object" in s and s[0] == '<' and s[-1] == '>':
            pos = s.find(" object")
            s = "<" + s[1:pos].replace("__main__.", "") + ">"
        return s.replace('"', "'")
    except:
        return "<%s>" % str(obj.__class__).replace('"', "'")

# _________________________________________________________________________________________
#
#   shorten_filenames_encoder
# ________________________________________________________________________________________


def shorten_filenames_encoder(obj, n_levels=2):
    """
    Convert a set of parameters into a string
        Paths with > N levels of nested-ness are truncated
    """

    #
    #   if < 0, nest by 2
    #
    if n_levels < 0:
        desired_len = -n_levels
        prev_encoded_len = 0
        #
        #   try more and more nestedness up to 9 if that fits inside desired length
        #       stop when increasing nestedness makes no difference
        #
        for nestedness in range(1, 20):
            res = shorten_filenames_encoder(obj, nestedness)
            if len(res) > desired_len or "..." not in res:
                break
            prev_encoded_len = len(res)
        desired_len = max(4, desired_len - 5)
        offset = len(res) - desired_len
        if offset < 0:
            return res
        return "<???> " + res[offset:]

    #
    #   Recurse into lists and tuples
    #
    if non_str_sequence(obj):
        return "[%s]" % ", ".join(map(shorten_filenames_encoder, obj, [n_levels] * len(obj)))

    #
    #   Only shorten strings
    #
    if not isinstance(obj, path_str_type):
        return ignore_unknown_encoder(obj)

    #
    #   level = 0 means return full absolute path
    #
    if not n_levels:
        return ignore_unknown_encoder(os.path.abspath(obj))

    #
    # Shorten both relative and absolute (full) paths
    #

    # if within bounds, return that
    if obj[1:].count('/') < n_levels:
        return ignore_unknown_encoder(obj)

    # use relative path if that has <= nested level
    rel_path = os.path.relpath(obj)
    if rel_path.count('/') <= n_levels:
        #print >>sys.stderr, "relative path only one nested level"
        return ignore_unknown_encoder(rel_path)

    # get last N nested levels
    # print >>sys.stderr, "full path last N nested level"
    return ignore_unknown_encoder(get_nth_nested_level_of_path(obj, n_levels))


# _________________________________________________________________________________________
#
#   get_tasks_filename_globs_in_nested_sequence
#
# ________________________________________________________________________________________
glob_letters = set('*[]?')


def is_glob(s):
    """Check whether 's' contains ANY of glob chars"""
    return len(glob_letters.intersection(s)) > 0


# _________________________________________________________________________________________
#
#   get_nested_tasks_or_globs
#
# ________________________________________________________________________________________
def get_nested_tasks_or_globs(p, treat_strings_as_tasks=False, runtime_data_names=None, tasks=None, globs=None):
    """
    Get any tasks or globs which are within parameter
        tasks are returned as functions or function names
    """
    #
    # create storage if this is not a recursive call
    #
    if globs is None:
        runtime_data_names, tasks, globs = set(), list(), set()

    #
    #   task function
    #
    if isinstance(p, Callable):
        tasks.append(p)
    elif p.__class__.__name__ == 'Task' or p.__class__.__name__ == 'Pipeline':
        tasks.append(p)
    elif isinstance(p, runtime_parameter):
        runtime_data_names.add(p)

    #
    #   output_from treats all arguments as tasks or task names
    #
    elif isinstance(p, output_from):
        for pp in p.args:
            get_nested_tasks_or_globs(
                pp, True, runtime_data_names, tasks, globs)

    elif isinstance(p, path_str_type):
        if treat_strings_as_tasks:
            tasks.append(p)
        elif is_glob(p):
            globs.add(p)

    elif non_str_sequence(p):
        for pp in p:
            get_nested_tasks_or_globs(
                pp, treat_strings_as_tasks, runtime_data_names, tasks, globs)
    return tasks, globs, runtime_data_names

# _________________________________________________________________________________________
#
#   replace_placeholders_with_tasks_in_input_params
#
# ________________________________________________________________________________________


def replace_placeholders_with_tasks_in_input_params(p, func_or_name_to_task, treat_strings_as_tasks=False):
    """
    Replaces task functions or task name (strings) with the tasks they represent
    Also replaces Tasks and Pipelines with the correct Tasks
    func_or_name_to_task are a dictionary of function and task names to tasks

    """
    if p.__class__.__name__ == 'Pipeline':
        return func_or_name_to_task["PIPELINE=%s=PIPELINE" % p.name]

    if p.__class__.__name__ == 'Task' and p in func_or_name_to_task:
        return func_or_name_to_task[p]

    #
    # Expand globs or tasks as a list only if they are top level
    #
    if isinstance(p, Callable):
        # if type(p) == types.FunctionType:
        return func_or_name_to_task[p]

    #
    #   output_from treats all arguments as tasks or task names
    #
    if isinstance(p, output_from):
        if len(p.args) == 1:
            return replace_placeholders_with_tasks_in_input_params(p.args[0], func_or_name_to_task, True)
        else:
            return [replace_placeholders_with_tasks_in_input_params(pp, func_or_name_to_task, True) for pp in p.args]

    #
    # strings become tasks if treat_strings_as_tasks
    #
    if isinstance(p, path_str_type):
        if treat_strings_as_tasks:
            return func_or_name_to_task[p]
        return p

    #
    # No conversions within dictionaries
    #
    if isinstance(p, dict):
        return p

    #
    # Other sequences are recursed down
    #
    elif non_str_sequence(p):
        l = list()
        for pp in p:

            #
            #   To be intuitive:
            #   arguments wrapped by output_from are always treated "in-line"
            #           e.g. 1, output_from("a") => 1, task_a
            #           e.g. 1, output_from("a", 2) => 1, task_a, 2
            #
            if isinstance(pp, output_from):
                if len(pp.args) > 1:
                    l.extend(tuple(replace_placeholders_with_tasks_in_input_params(
                        pp, func_or_name_to_task, True)))
                elif len(pp.args) == 1:
                    l.append(replace_placeholders_with_tasks_in_input_params(
                        pp.args[0], func_or_name_to_task, True))
                # else len(pp.args) == 0 !! do nothing

            else:
                l.append(replace_placeholders_with_tasks_in_input_params(
                    pp, func_or_name_to_task, treat_strings_as_tasks))
        return type(p)(l)

    #
    # No expansions of non-string/non-sequences
    #
    else:
        return p

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   compiling regular expressions

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
# _________________________________________________________________________________________

#   suffix

# _________________________________________________________________________________________


class suffix(object):
    def __init__(self, *args):
        self.args = args

    def __repr__(self, *args):
        return 'suffix%r' % (self.args,)

# _________________________________________________________________________________________

#   regex

# _________________________________________________________________________________________


class regex(object):
    def __init__(self, *args):
        self.args = args

    def __repr__(self, *args):
        return 'regex%r' % (self.args,)

# _________________________________________________________________________________________

#   regex

# _________________________________________________________________________________________


class formatter(object):
    def __init__(self, *args):
        self.args = args

    def __repr__(self, *args):
        return 'formatter%r' % (self.args,)

# _________________________________________________________________________________________

#   wrap_exception_as_string

# _________________________________________________________________________________________


def wrap_exception_as_string():
    """
    return exception as string to be rethrown
    """
    exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
    msg = "%s.%s" % (exceptionType.__module__, exceptionType.__name__)
    exception_value = str(exceptionValue)
    if len(exception_value):
        return msg + ": (%s)" % exception_value
    return msg


# _________________________________________________________________________________________

#   compile_formatter

# _________________________________________________________________________________________
def compile_formatter(enclosing_task, formatter_obj, error_object, descriptor_string):
    """
    Given list of [string|None]
    Return compiled regular expressions.
    """

    compiled_regexes = []
    for ss in formatter_obj.args:
        # ignore None
        if ss is None:
            compiled_regexes.append(None)
            continue

        formatter_args = str(formatter_obj.args)[1:-1]
        # regular expression should be strings
        if not isinstance(ss, path_str_type):
            raise error_object(enclosing_task, ("{descriptor_string}: "
                                                "formatter({formatter_args}) is malformed\n"
                                                "formatter(...) should only be used to wrap "
                                                'regular expression strings or None (not "{ss}")')
                               .format(descriptor_string=descriptor_string,
                                       formatter_args=formatter_args,
                                       ss=ss)
                               )

        try:
            compiled_regexes.append(re.compile(ss))
        except:
            raise error_object(enclosing_task, ("{descriptor_string}: "
                                                "in formatter({formatter_args}) \n"
                                                'regular expression "{ss}" is malformed\n'
                                                "[{except_str}]")
                               .format(descriptor_string=descriptor_string,
                                       formatter_args=formatter_args,
                                       ss=ss,
                                       except_str=wrap_exception_as_string())
                               )
    return compiled_regexes


# _________________________________________________________________________________________

#   compile_regex

# _________________________________________________________________________________________
def compile_regex(enclosing_task, regex, error_object, descriptor_string, regex_object_name="regex"):
    """
    throw error unless regular expression compiles
    """
    if not len(regex.args) or len(regex.args) > 1 or not isinstance(regex.args[0], path_str_type):
        regex_str = str(regex.args)
        if len(regex.args) > 1:
            regex_str = regex_str[1:-1]
        elif len(regex.args) == 0:
            regex_str = ''
        else:
            regex_str = regex_str
        raise error_object(enclosing_task, ("{descriptor_string}: "
                                            "{regex_object_name}({regex_str}) is malformed\n"
                                            "{regex_object_name}(...) should only be used to wrap a single regular expression string")
                           .format(descriptor_string=descriptor_string,
                                   regex_str=regex_str,
                                   regex_object_name=regex_object_name)
                           )
    try:
        matching_regex = re.compile(regex.args[0])
        return matching_regex
    except:
        raise error_object(enclosing_task, ("{descriptor_string}: "
                                            "regular expression {regex_object_name}('{regex_str}') is malformed\n"
                                            "[{except_str}]")
                           .format(descriptor_string=descriptor_string,
                                   regex_object_name=regex_object_name,
                                   regex_str=regex.args[0],
                                   except_str=wrap_exception_as_string())
                           )

# _________________________________________________________________________________________

#   compile_suffix

# _________________________________________________________________________________________


def compile_suffix(enclosing_task, regex, error_object, descriptor_string):
    """
    throw error unless regular expression compiles
    """
    if not len(regex.args):
        raise error_object(enclosing_task, "%s: " % descriptor_string +
                           "suffix() is malformed.\n" +
                           "suffix(...) should be used to wrap a string matching the suffices of file names")
    if len(regex.args) > 1 or not isinstance(regex.args[0], path_str_type):
        raise error_object(enclosing_task, "%s: " % descriptor_string +
                           "suffix('%s') is malformed.\n" % (regex.args,) +
                           "suffix(...) should only be used to wrap a single string matching the suffices of file names")
    try:
        matching_regex = re.compile("(.*)" + re.escape(regex.args[0]) + "$")
        return matching_regex
    except:
        raise error_object(enclosing_task, "%s: " % descriptor_string +
                           "suffix('%s') is somehow malformed\n" % regex.args[0] +
                           "[%s]" % wrap_exception_as_string())

# _________________________________________________________________________________________

#   check_parallel_parameters

# _________________________________________________________________________________________


def check_parallel_parameters(enclosing_task, params, error_object):
    """
    Helper function for @files
    Checks format of parameters and
    whether there are input and output files specified for each job
    """
    if not len(params):
        raise Exception("@parallel parameters is empty.")

    for job_param in params:
        if isinstance(job_param, path_str_type):
            message = ("Wrong syntax for @parallel.\n"
                       "@parallel(%s)\n" % ignore_unknown_encoder(params) +
                       "If you are supplying parameters for a task "
                       "running as a single job, "
                       "either don't put enclosing brackets at all (with each parameter "
                       "separated by commas) or enclose all parameters as a nested list of "
                       "lists, e.g. [['input', 'output' ...]]. "
                       )
            raise error_object(enclosing_task, message)


# _________________________________________________________________________________________

#   check_files_io_parameters

# _________________________________________________________________________________________
def check_files_io_parameters(enclosing_task, params, error_object):
    """
    Helper function for @files
    Checks format of parameters and
    whether there are input and output files specified for each job
    """
    # if not len(params):
    #    raise Exception("@files I/O parameters is empty.")

    try:
        for job_param in params:
            if isinstance(job_param, path_str_type):
                raise TypeError

            if len(job_param) < 1:
                raise error_object(enclosing_task, "Missing input files for job " +
                                   ignore_unknown_encoder(job_param))
            if len(job_param) < 2:
                raise error_object(enclosing_task, "Missing output files for job " +
                                   ignore_unknown_encoder(job_param))
            # if len(get_strings_in_flattened_sequence(job_param[0:2])) == 0:
            #    raise error_object(enclosing_task, "Input or output file parameters should "
            #                                        "contain at least one or more file names strings. "
            #                                        "Consider using @parallel if you are not using files. " +
            #                                        ignore_unknown_encoder(job_param))
    except TypeError:
        #
        # job_param was not a list
        #
        message = ("Wrong syntax for @files.\n@files(%s)\n" % ignore_unknown_encoder(params) +
                   "If you are supplying parameters for a task "
                   "running as a single job, "
                   "either don't put enclosing brackets at all (with each parameter "
                   "separated by commas) or enclose all parameters as a nested list of "
                   "lists, e.g. [['input', 'output' ...]]. "
                   )
        raise error_object(enclosing_task, message)

# _________________________________________________________________________________________
#
#   expand_nested_tasks_or_globs
#
# ________________________________________________________________________________________


def expand_nested_tasks_or_globs(p, tasksglobs_to_filenames):
    """
    Expand globs and tasks "in-line", unless they are the top level, in which case turn
    it into a list

    N.B. Globs are only expanded if they are in tasksglobs_to_filenames
         This function is called for @split descriptors which leave output globs untouched
         for clarity. Thanks to Noah Spies for spotting this!
    """

    #
    # Expand globs or tasks as a list only if they are top level
    #
    if ((isinstance(p, path_str_type) and is_glob(p) and p in tasksglobs_to_filenames) or
            p.__class__.__name__ == 'Task' or
            isinstance(p, runtime_parameter)):
        return tasksglobs_to_filenames[p]

    #
    # No expansions of strings and dictionaries
    #
    if isinstance(p, (path_str_type, dict)):
        return p

    #
    # Other sequences are recursed down
    #
    elif non_str_sequence(p):
        l = list()
        for pp in p:
            if (isinstance(pp, path_str_type) and pp in tasksglobs_to_filenames):
                l.extend(tasksglobs_to_filenames[pp])
            elif pp.__class__.__name__ == 'Task' or isinstance(pp, runtime_parameter):
                files = tasksglobs_to_filenames[pp]
                # task may have produced a single output: in which case append
                if non_str_sequence(files):
                    l.extend(files)
                else:
                    l.append(files)
            else:
                l.append(expand_nested_tasks_or_globs(
                    pp, tasksglobs_to_filenames))
        return type(p)(l)

    #
    # No expansions of non-string/non-sequences
    #
    else:
        return p


# _________________________________________________________________________________________

#   get_parsed_arguments_str_for_errors

#       helper funciton for parse_task_arguments()

# _________________________________________________________________________________________
def get_parsed_arguments_str_for_errors(task_description, bad_arg_str, unnamed_result_strs, named_result_strs):
    """
    Helper function for parse_task_arguments
        Prints out offending argument (bad_arg_str) in the context of already parsed
        arguments so that we can quickly figure out where the error is coming from
    """
    indent = task_description.find("(") + 1
    parsed_arg_str = ", ".join(unnamed_result_strs + named_result_strs)
    # make function names clearer in arg list
    parsed_arg_str = re.sub(
        r"<function (\w+) at 0x[0-9a-f]+>", r"\1", parsed_arg_str)
    return "\n" + task_description % (parsed_arg_str + ", ...\n" +
                                      # mark out problem
                                      (" " * (indent-5 if indent - 5 > 0 else 0)) + "===> " +
                                      bad_arg_str)


# _________________________________________________________________________________________

#   parse_task_arguments

# _________________________________________________________________________________________
def parse_task_arguments(orig_unnamed_arguments, orig_named_arguments, expected_arguments, task_description):
    """
    Parse arguments parsed into decorators or Pipeline.transform etc.
        Special handling for optional arguments in the middle of argument list
            1) @product
                can have (input, filter, input1, filter1, input2, filter2....)
            2) @transform, @subdivide, @collate, @product, @combinatorics which have
                    (..., [add_inputs(...)|inputs(...)],...)
                    or ([add_inputs=...|replace_inputs=...])
                    or ([add_inputs=add_inputs(...)|replace_inputs=inputs(...)])
        Special handling for variable number of arguments at the end of the argument list
            which all become "extras"

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    #
    #   N.B. Missing non-mandatory arguments are returned as an empty list
    #
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

                """
    results = {}
    unnamed_arguments = list(orig_unnamed_arguments)
    named_arguments = dict(orig_named_arguments)
    # parsed results in string form for error messages
    unnamed_result_strs = []
    named_result_strs = []

    def parse_add_inputs_args(parsed_arg, input_type, arg_name, modify_inputs_mode, result_strs):
        """
        Parse arguments for add_inputs and replace_inputs, i.e. 'inputs()' and 'add_inputs()'
            input_type =inputs|add_inputs
            arg_name = replace_inputs | add_inputs
            modify_inputs_mode = t_extra_inputs.REPLACE_INPUTS| t_extra_inputs.ADD_TO_INPUTS
        """
        results["modify_inputs_mode"] = modify_inputs_mode
        if input_type == inputs:
            # inputs() only takes a single argument. Throw error otherwise
            if len(parsed_arg.args) != 1:
                err_msg = "inputs() expects a single argument:\n%s" % (
                    get_parsed_arguments_str_for_errors(task_description,   # bad arg in context of parsed
                                                        "%s%r" % (input_type.__name__, tuple(
                                                            parsed_arg.args)),
                                                        unnamed_result_strs,
                                                        named_result_strs))
                raise error_inputs_multiple_args(err_msg)
            # unpack add_inputs / inputs and save results
            results["modify_inputs"] = parsed_arg.args[0]
        else:
            results["modify_inputs"] = parsed_arg.args
        result_strs.append("%s=%r" % (arg_name, parsed_arg.args))

    def check_argument_type(arg_name, parsed_arg, argument_types):
        """
        check if parsed_arg is right type
        """
        if argument_types and not isinstance(parsed_arg, argument_types):
            err_msg = ("The '%s' argument should be %s:\n%s" %
                       (arg_name,                                                  # argument name
                        # type names
                        " or ".join("%s" %
                                    tt.__name__ for tt in argument_types),
                        get_parsed_arguments_str_for_errors(task_description,       # bad arg in context of parsed
                                                            "%s = %r" % (
                                                                arg_name, parsed_arg),
                                                            unnamed_result_strs, named_result_strs)))
            #print (err_msg, file=sys.stderr)
            raise TypeError(err_msg)

        return parsed_arg

    def parse_argument(arg_name, expected_arguments, unnamed_arguments, named_arguments,
                       results, task_description, mandatory, argument_types=None):
        """
        All missing, non-mandatory are empty list
        """

        # ignore if not on list
        if not arg_name in expected_arguments:
            return

        #
        # look among unnamed arguments first
        #
        if len(unnamed_arguments):
            # check correct type
            parsed_arg = check_argument_type(
                arg_name, unnamed_arguments[0], argument_types)
            # save parsed results
            results[arg_name] = parsed_arg
            unnamed_result_strs.append("%s=%r" % (arg_name, parsed_arg))
            del unnamed_arguments[0]

        #
        # then named
        #
        elif arg_name in named_arguments:
            #
            # check correct type
            #
            parsed_arg = check_argument_type(
                arg_name, named_arguments[arg_name], argument_types)
            #
            #   Save parsed results
            #
            results[arg_name] = parsed_arg
            named_result_strs.append("%s=%r" % (arg_name, parsed_arg))
            del named_arguments[arg_name]

        #
        # complain or ignore missing?
        #
        else:
            if mandatory:
                err_msg = "Missing '%s' argument:\n%s" % (
                    arg_name,                                               # argument name
                    get_parsed_arguments_str_for_errors(task_description,   # bad arg in
                                                        arg_name + " = ???",  # context of parsed
                                                        unnamed_result_strs,
                                                        named_result_strs))
                #print (err_msg, file=sys.stderr)
                raise error_missing_args(err_msg)

            #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            #
            #   N.B. Missing non-mandatory arguments are returned as an empty list
            #
            #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            else:
                results[arg_name] = []

    #
    #   Missing input is empty list
    #
    parse_argument('input', expected_arguments, unnamed_arguments,
                   named_arguments, results, task_description, mandatory=False)

    #
    #   filter is mandatory if expected
    #
    parse_argument('filter', expected_arguments, unnamed_arguments, named_arguments, results,
                   task_description, mandatory=True, argument_types=(formatter, regex, suffix))

    if "filter" in results:
        if isinstance(results["filter"], suffix):
            parse_argument("output_dir", expected_arguments, [
            ], named_arguments, results, task_description, mandatory=False)

    #
    #   inputN
    #
    if 'inputN' in expected_arguments:
        #
        # put already parsed input and filter into the list
        #
        results["input"] = [results["input"]]
        results["filter"] = [results["filter"]]
        cnt_inputN = 1
        #
        #   parse argument pairs at a time, so long as the second argument is a formatter
        #
        while len(unnamed_arguments) >= 2 and isinstance(unnamed_arguments[1], formatter):
            filter_name = "filter%d" % (cnt_inputN + 1)
            input_name = "input%d" % (cnt_inputN + 1)
            unnamed_result_strs.append("%s=%r" % (
                input_name, unnamed_arguments[0]))
            unnamed_result_strs.append("%s=%r" % (
                filter_name, unnamed_arguments[1]))
            results["input"].append(unnamed_arguments[0])
            results["filter"].append(unnamed_arguments[1])
            cnt_inputN += 1
            del unnamed_arguments[0:2]

        #
        #   parse named arguments while there is a filter2, filter3 etc.
        #
        filter_name = "filter%d" % (cnt_inputN + 1)
        input_name = "input%d" % (cnt_inputN + 1)
        while filter_name in named_arguments:
            results["filter"].append(named_arguments[filter_name])
            named_result_strs.append("%s=%r" % (
                filter_name, named_arguments[filter_name]))
            del named_arguments[filter_name]
            #   parse input2, input3 or leave blank as empty list
            if input_name in named_arguments:
                results["input"].append(named_arguments[input_name])
                named_result_strs.append("%s=%r" % (
                    input_name, named_arguments[input_name]))
                del named_arguments[input_name]
            else:
                results["input"].append([])
            cnt_inputN += 1
            filter_name = "filter%d" % (cnt_inputN + 1)
            input_name = "input%d" % (cnt_inputN + 1)

    #
    #   tuple size is int and mandatory if exists
    #
    parse_argument('tuple_size', expected_arguments, unnamed_arguments, named_arguments,
                   results, task_description, mandatory=True, argument_types=(int,))

    #
    #   add_inputs / inputs are optional
    #
    if 'modify_inputs' in expected_arguments:
        results["modify_inputs_mode"] = t_extra_inputs.KEEP_INPUTS
        results["modify_inputs"] = None
        parse_add_inputs = ((inputs, "inputs", "replace_inputs", t_extra_inputs.REPLACE_INPUTS),
                            (add_inputs, "add_inputs", "add_inputs", t_extra_inputs.ADD_TO_INPUTS))

        if len(unnamed_arguments):
            #
            #   Is add_inputs or inputs in unnamed arguments?
            #       Parse out contents and save in results["replace_inputs"] or results["add_inputs"]
            #
            for input_type, input_type_name, arg_name, modify_inputs_mode in parse_add_inputs:
                parsed_arg = unnamed_arguments[0]
                if isinstance(parsed_arg, input_type):
                    parse_add_inputs_args(
                        parsed_arg, input_type, arg_name, modify_inputs_mode, unnamed_result_strs)
                    del unnamed_arguments[0]
                    break
        #
        #   Otherwise is add_inputs or inputs in named arguments?
        #       Parse out contents only if necessary and save in results["replace_inputs"] or results["add_inputs"]
        #
        if results["modify_inputs_mode"] == t_extra_inputs.KEEP_INPUTS:
            for input_type, input_type_name, arg_name, modify_inputs_mode in parse_add_inputs:
                if arg_name in named_arguments:
                    parsed_arg = named_arguments[arg_name]
                    if isinstance(parsed_arg, input_type):
                        parse_add_inputs_args(
                            parsed_arg, input_type, arg_name, modify_inputs_mode, named_result_strs)
                    else:
                        results["modify_inputs"] = parsed_arg
                    results["modify_inputs_mode"] = modify_inputs_mode
                    del named_arguments[arg_name]
                    break

    #
    #   output is mandatory if exists
    #
    parse_argument('output', expected_arguments, unnamed_arguments,
                   named_arguments, results, task_description, mandatory=True)

    #
    #   extras is mandatory if exists
    #
    if 'extras' in expected_arguments:
        results['extras'] = []
        results['named_extras'] = {}
        if len(unnamed_arguments):
            # move list to results: remember python does shallow copies of lists
            results['extras'] = unnamed_arguments
            unnamed_result_strs.append("%s=%r" % ("extras", unnamed_arguments))
            unnamed_arguments = []
            #del unnamed_arguments[:]
        elif 'extras' in named_arguments:
            # Named extras only
            if isinstance(named_arguments['extras'], dict):
                results["named_extras"] = named_arguments['extras']
            # Unnamed extras only
            elif isinstance(named_arguments['extras'], list):
                results["extras"] = named_arguments['extras']
            # Wrong type: blow up
            else:
                err_msg = ("The extras paramter must be either a list of values\nor a dictionary of named parameter values:\n%s" %
                           get_parsed_arguments_str_for_errors(task_description,
                                                               "extras=%r" % (
                                                                   named_arguments['extras'],),
                                                               unnamed_result_strs,
                                                               named_result_strs))
                raise error_extras_wrong_type(err_msg)

            named_result_strs.append("%s=%r" % (
                "extras", named_arguments['extras']))
            del named_arguments['extras']

    if len(unnamed_arguments):
        err_msg = ("Too many unnamed arguments leftover:\n%s" %
                   get_parsed_arguments_str_for_errors(task_description,       # bad arg in context of parsed
                                                       (", ".join(("%r" % a)
                                                                  for a in unnamed_arguments)),
                                                       unnamed_result_strs, named_result_strs))
        #print (err_msg, file=sys.stderr)
        raise error_too_many_args(err_msg)
    if len(named_arguments):
        err_msg = ("Duplicate, conflicting or unrecognised arguments:\n%s" %
                   get_parsed_arguments_str_for_errors(task_description,       # bad arg in context of parsed
                                                       ", ".join("%s=%r" % (
                                                           k, v) for k, v in named_arguments.items()),
                                                       unnamed_result_strs, named_result_strs))
        #print (err_msg, file=sys.stderr)
        raise error_too_many_args(err_msg)

    return results

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   special markers used by @files_re

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


class combine(object):
    def __init__(self, *args):
        self.args = args


class output_from(object):
    def __init__(self, *args):
        self.args = args

    def __repr__(self, *args):
        return 'output_from%r' % (self.args,)


class runtime_parameter(object):
    def __init__(self, *args):
        if len(args) != 1 or not isinstance(args[0], path_str_type):
            raise Exception(
                "runtime_parameter takes the name of the run time parameter as a single string")
        self.args = args


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Unit Testing code in test/test_ruffus_utility.py


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888



#!/usr/bin/env python
from __future__ import print_function
import shutil
import unittest
from subprocess import *
from ruffus import pipeline_run, pipeline_printout_graph, originate, split, transform, \
    subdivide, formatter, Pipeline
import sys
"""

    test_split_subdivide_checkpointing.py


"""
import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Each time the pipeline is FORCED to rerun,
#       More files are created for each task


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


@originate([tempdir + 'start'])
def make_start(outfile):
    """
    -> start
    """
    open(outfile, 'w').close()


@split(make_start, tempdir + '*.split')
def split_start(infiles, outfiles):
    """
    -> XXX.split
        where XXX = 0 .. N,
              N   = previous N + 1
    """

    # split always runs exactly one job (unlike @subdivide)
    # So it implicitly combines all its inputs before running and generating multiple output
    # @originate generates multiple output so the input for @split is a list...
    infile = infiles[0]

    # clean up previous
    for f in outfiles:
        os.unlink(f)

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    #
    #   Create more files than the previous invocation
    #
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    n_to_produce = len(outfiles) + 1
    for i in range(n_to_produce):
        f = '{}{}.split'.format(tempdir, i)
        open(f, 'a').close()


@subdivide(split_start, formatter(), tempdir + '{basename[0]}_*.subdivided', tempdir + '{basename[0]}')
def subdivide_start(infile, outfiles, infile_basename):
    # cleanup existing
    for f in outfiles:
        os.unlink(f)

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    #
    #   Create more files than the previous invocation
    #
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    n_to_produce = len(outfiles) + 1
    for i in range(n_to_produce):
        open('{}_{}.subdivided'.format(infile_basename, i), 'a').close()


class Test_ruffus(unittest.TestCase):

    def tearDown(self):
        # only tear down if not throw exception so we can debug?
        try:
            shutil.rmtree(tempdir)
        except:
            pass

    def setUp(self):
        try:
            shutil.rmtree(tempdir)
        except:
            pass
        os.makedirs(tempdir)

        #
        #   check graphviz exists for turning dot files into jpg, svg etc
        #
        try:
            process = Popen("echo what | dot", stdout=PIPE,
                            stderr=STDOUT, shell=True)
            output, unused_err = process.communicate()
            retcode = process.poll()
            if retcode:
                raise CalledProcessError(
                    retcode, "echo what | dot", output=output)
        except CalledProcessError as err:
            output_str = str(err.output)
            if "No such file or directory" in output_str or "not found" in output_str or "Unable to access jarfile" in output_str:
                self.graph_viz_present = False
                return
        self.graph_viz_present = True

    def test_ruffus(self):

        print("     Run pipeline normally...")
        if self.graph_viz_present:
            pipeline_printout_graph(tempdir + "flowchart.dot", pipeline="main")
            pipeline_printout_graph(tempdir + "flowchart.jpg",
                                    target_tasks=[subdivide_start],
                                    forcedtorun_tasks=[split_start],
                                    no_key_legend=True)
            pipeline_printout_graph(
                tempdir + "flowchart.svg", no_key_legend=False, pipeline="main")
            # Unknown format
            try:
                pipeline_printout_graph(
                    tempdir + "flowchart.unknown", no_key_legend=False, pipeline="main")
                raise Exception(
                    "Failed to throw exception for pipeline_printout_graph unknown extension ")
            except CalledProcessError as err:
                pass
            pipeline_printout_graph(
                tempdir + "flowchart.unknown", "svg", no_key_legend=False, pipeline="main")

        else:
            pipeline_printout_graph(tempdir + "flowchart.dot",
                                    target_tasks=[subdivide_start],
                                    forcedtorun_tasks=[split_start],
                                    no_key_legend=True,
                                    pipeline="main")

    def test_newstyle_ruffus(self):

        print("     Run pipeline normally...")
        test_pipeline = Pipeline("test")
        test_pipeline.originate(make_start, [tempdir + 'start'])

        test_pipeline.split(split_start, make_start, tempdir + '*.split')

        test_pipeline.subdivide(subdivide_start, split_start, formatter(
        ), tempdir + '{basename[0]}_*.subdivided', tempdir + '{basename[0]}')
        if self.graph_viz_present:
            test_pipeline.printout_graph(tempdir + "flowchart.dot")
            test_pipeline.printout_graph(tempdir + "flowchart.jpg",
                                         target_tasks=[subdivide_start],
                                         forcedtorun_tasks=[split_start],
                                         no_key_legend=True)
            test_pipeline.printout_graph(
                tempdir + "flowchart.svg", no_key_legend=False)
            # Unknown format
            try:
                test_pipeline.printout_graph(
                    tempdir + "flowchart.unknown", no_key_legend=False)
                raise Exception(
                    "Failed to throw exception for test_pipeline.printout_graph unknown extension ")
            except CalledProcessError as err:
                pass
            test_pipeline.printout_graph(
                tempdir + "flowchart.unknown", "svg", no_key_legend=False)

        else:
            test_pipeline.printout_graph(tempdir + "flowchart.dot",
                                         target_tasks=[subdivide_start],
                                         forcedtorun_tasks=[split_start],
                                         no_key_legend=True)


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python

from __future__ import print_function
import ruffus
import ruffus.drmaa_wrapper
import os
import sys
import time


try:
    import drmaa
    HAVE_DRMAA = True
    DRMAA_SESSION = drmaa.Session()
    DRMAA_SESSION.initialize()

except ImportError:
    HAVE_DRMAA = False


logger, logger_mutex = ruffus.cmdline.setup_logging("me", "test.log", 1)

exe_path = os.path.split(os.path.abspath(sys.argv[0]))[0]
sys.path.insert(0, os.path.abspath(os.path.join(exe_path, "..", "..")))

workdir = 'tmp_test_job_history_with_exceptions'
# sub-1s resolution in system?
one_second_per_job = None
throw_exception = False


def run_job(infile, outfile, **kwargs):
    if not kwargs.get("run_locally", False):
        kwargs["drmaa_session"] = DRMAA_SESSION

    print("%s start to run " % infile)
    stdout, stderr = ruffus.drmaa_wrapper.run_job(
        cmd_str="sleep 100 && hostname > {}".format(os.path.abspath(outfile)),
        verbose=1,
        local_echo=False,
        logger=logger,
        **kwargs)
    print("%s completed" % infile)


@ruffus.mkdir(workdir)
@ruffus.originate([workdir + "/" + prefix + "_name.tmp1" for prefix in "abcdefghijk"])
def generate_initial_files1(on):
    with open(on, 'w') as outfile:
        pass


@ruffus.transform(generate_initial_files1,
                  ruffus.suffix(".tmp1"), ".tmp2")
def test_task2(infile, outfile):
    run_job(infile, outfile)


@ruffus.transform(test_task2, ruffus.suffix(".tmp2"), ".tmp3")
def test_task3(infile, outfile):
    run_job(infile, outfile)


def cleanup_tmpdir():
    os.system('rm -f %s %s' %
              (os.path.join(workdir, '*'), ruffus.RUFFUS_HISTORY_FILE))


def do_main():
    print("Press Ctrl-C Now!!", file=sys.stdout)
    sys.stdout.flush()
    time.sleep(2)
    print("Start....", file=sys.stdout)
    sys.stdout.flush()
    ruffus.pipeline_run(verbose=5,
                        pool_manager="gevent",
                        multiprocess=5,
                        pipeline="main")
    print("too late!!", file=sys.stdout)
    sys.stdout.flush()
    cleanup_tmpdir()


do_main()



################################################################################
#
#   print_dependencies.py
#
#
#   Copyright (c) 10/9/2009 Leo Goodstadt
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#################################################################################
from collections import defaultdict
from .adjacent_pairs_iterate import adjacent_pairs_iterate
"""
    print_dependencies.py

        provides support for dependency trees

"""

#
#   Number of pre-canned colour schemes
#
CNT_COLOUR_SCHEMES = 8


try:
    from StringIO import StringIO
except:
    from io import StringIO


def _get_name(node):
    """
    Get name for node
        use display_name or _name
    """
    if hasattr(node, "display_name"):
        return node.display_name
    elif hasattr(node, "_name"):
        return node._name
    else:
        raise Exception(
            "Unknown node type [%s] has neither _name or display_name" % str(node))


def attributes_to_str(attributes, name):
    """
    helper function for dot format
    turns dictionary into a=b, c=d...
    """

    # remove ugly __main__. qualifier
    name = name.replace("__main__.", "")

    # if a label is specified, that overrides the node name
    if "label" not in attributes:
        attributes["label"] = name.replace(
            "  before ", "\\nbefore ").replace(",  ", ",\n")

    # remove any quotes
    if attributes["label"][0] == '<':
        attributes["label"] = attributes["label"][1:-1]
        html_label = True
    else:
        html_label = False
        if attributes["label"][0] == '"':
            attributes["label"] = attributes["label"][1:-1]

    # add suffix / prefix
    if "label_prefix" in attributes:
        attributes["label"] = attributes["label_prefix"] + attributes["label"]
        del attributes["label_prefix"]
    if "label_suffix" in attributes:
        attributes["label"] = attributes["label"] + attributes["label_suffix"]
        del attributes["label_suffix"]

    # restore quotes
    if html_label:
        attributes["label"] = '<' + attributes["label"] + '>'
    else:
        attributes["label"] = '"' + attributes["label"] + '"'

    # support for html labels
    # if "<" in name and ">" in name:
    #    attributes["label"] = '<' + name + '>'
    # else:
    #    attributes["label"] = '"' + name + '"'

    return "[" + ", ".join("%s=%s" % (k, v) for k, v in sorted(attributes.items())) + "];\n"


# _________________________________________________________________________________________
#
#   get_arrow_str_for_legend_key
# _________________________________________________________________________________________
def get_arrow_str_for_legend_key(from_task_type, to_task_type, n1, n2, colour_scheme):
    """
    Get dot format for arrows inside legend key
    """
    if "Vicious cycle" in (from_task_type, to_task_type):
        return ("%s -> %s[color=%s, arrowtype=normal];\n" % (n1, n2, colour_scheme["Vicious cycle"]["linecolor"]) +
                "%s -> %s[color=%s, arrowtype=normal];\n" % (n2, n1, colour_scheme["Vicious cycle"]["linecolor"]))
    if from_task_type in ("Final target", "Task to run",
                          "Up-to-date task forced to rerun",
                          "Explicitly specified task"):
        return "%s -> %s[color=%s, arrowtype=normal];\n" % (n1, n2, colour_scheme["Task to run"]["linecolor"])
    elif from_task_type in ("Up-to-date task", "Down stream", "Up-to-date Final target"):
        return "%s -> %s[color=%s, arrowtype=normal];\n" % (n1, n2, colour_scheme["Up-to-date"]["linecolor"])
    #
    # shouldn't be here!!
    #
    else:
        return "%s -> %s[color=%s, arrowtype=normal];\n" % (n1, n2, colour_scheme["Up-to-date"]["linecolor"])


# _________________________________________________________________________________________
#
#   get_default_colour_scheme
# _________________________________________________________________________________________
def get_default_colour_scheme(default_colour_scheme_index=0):
    """
    A selection of default colour schemes "inspired" by entries in
        http://kuler.adobe.com/#create/fromacolor
    """

    if default_colour_scheme_index == 0:
        bluey_outline = '"#0044A0"'
        bluey = '"#EBF3FF"'
        greeny_outline = '"#006000"'
        greeny = '"#B8CC6E"'
        orangey = '"#EFA03B"'
        orangey_outline = greeny_outline
        ruddy = '"#FF3232"'
    elif default_colour_scheme_index == 1:
        bluey_outline = '"#000DDF"'
        bluey = 'transparent'
        greeny_outline = '"#4B8C2E"'
        greeny = '"#9ED983"'
        orangey = '"#D98100"'
        orangey_outline = '"#D9D911"'
        ruddy = '"#D93611"'
    elif default_colour_scheme_index == 2:
        bluey_outline = '"#4A64A5"'
        bluey = 'transparent'
        greeny_outline = '"#4A92A5"'
        greeny = '"#99D1C1"'
        orangey = '"#D2C24A"'
        orangey_outline = greeny_outline
        ruddy = '"#A54A64"'
    elif default_colour_scheme_index == 3:
        bluey_outline = '"#BFB5FF"'
        bluey = 'transparent'
        greeny_outline = '"#7D8A2E"'
        greeny = '"#C9D787"'
        orangey = '"#FFF1DC"'
        orangey_outline = greeny_outline
        ruddy = '"#FF3E68"'
    elif default_colour_scheme_index == 4:
        bluey_outline = '"#004460"'
        bluey = 'transparent'
        greeny_outline = '"#4B6000"'
        greeny = '"#B8CC6E"'
        orangey = '"#FFF0A3"'
        orangey_outline = greeny_outline
        ruddy = '"#F54F29"'
    elif default_colour_scheme_index == 5:
        bluey_outline = '"#1122FF"'
        bluey = '"#AABBFF"'
        greeny_outline = '"#007700"'
        greeny = '"#44FF44"'
        orangey = '"#EFA03B"'
        orangey_outline = '"#FFCC3B"'
        ruddy = '"#FF0000"'
    elif default_colour_scheme_index == 6:
        bluey_outline = '"#0044A0"'
        bluey = '"#EBF3FF"'
        greeny_outline = 'black'
        greeny = '"#6cb924"'
        orangey = '"#ece116"'
        orangey_outline = greeny_outline
        ruddy = '"#FF3232"'
    else:
        bluey_outline = '"#87BAE4"'
        bluey = 'transparent'
        greeny_outline = '"#87B379"'
        greeny = '"#D3FAE3"'
        orangey = '"#FDBA40"'
        orangey_outline = greeny_outline
        ruddy = '"#b9495e"'
    default_colour_scheme = defaultdict(dict)
    default_colour_scheme["Vicious cycle"]["linecolor"] = ruddy
    default_colour_scheme["Pipeline"]["fontcolor"] = ruddy
    default_colour_scheme["Key"]["fontcolor"] = "black"
    default_colour_scheme["Key"]["fillcolor"] = '"#F6F4F4"'
    default_colour_scheme["Task to run"]["linecolor"] = bluey_outline
    default_colour_scheme["Up-to-date"]["linecolor"] = "gray"
    default_colour_scheme["Final target"]["fillcolor"] = orangey
    default_colour_scheme["Final target"]["fontcolor"] = "black"
    default_colour_scheme["Final target"]["color"] = "black"
    default_colour_scheme["Final target"]["dashed"] = 0
    default_colour_scheme["Vicious cycle"]["fillcolor"] = ruddy
    default_colour_scheme["Vicious cycle"]["fontcolor"] = 'white'
    default_colour_scheme["Vicious cycle"]["color"] = "white"
    default_colour_scheme["Vicious cycle"]["dashed"] = 0
    default_colour_scheme["Up-to-date task"]["fillcolor"] = greeny
    default_colour_scheme["Up-to-date task"]["fontcolor"] = greeny_outline
    default_colour_scheme["Up-to-date task"]["color"] = greeny_outline
    default_colour_scheme["Up-to-date task"]["dashed"] = 0
    default_colour_scheme["Down stream"]["fillcolor"] = "white"
    default_colour_scheme["Down stream"]["fontcolor"] = "gray"
    default_colour_scheme["Down stream"]["color"] = "gray"
    default_colour_scheme["Down stream"]["dashed"] = 0
    default_colour_scheme["Explicitly specified task"]["fillcolor"] = "transparent"
    default_colour_scheme["Explicitly specified task"]["fontcolor"] = "black"
    default_colour_scheme["Explicitly specified task"]["color"] = "black"
    default_colour_scheme["Explicitly specified task"]["dashed"] = 0
    default_colour_scheme["Task to run"]["fillcolor"] = bluey
    default_colour_scheme["Task to run"]["fontcolor"] = bluey_outline
    default_colour_scheme["Task to run"]["color"] = bluey_outline
    default_colour_scheme["Task to run"]["dashed"] = 0
    default_colour_scheme["Up-to-date task forced to rerun"]["fillcolor"] = 'transparent'
    default_colour_scheme["Up-to-date task forced to rerun"]["fontcolor"] = bluey_outline
    default_colour_scheme["Up-to-date task forced to rerun"]["color"] = bluey_outline
    default_colour_scheme["Up-to-date task forced to rerun"]["dashed"] = 1
    default_colour_scheme["Up-to-date Final target"]["fillcolor"] = orangey
    default_colour_scheme["Up-to-date Final target"]["fontcolor"] = orangey_outline
    default_colour_scheme["Up-to-date Final target"]["color"] = orangey_outline
    default_colour_scheme["Up-to-date Final target"]["dashed"] = 0

    if default_colour_scheme_index == 6:
        default_colour_scheme["Vicious cycle"]["fontcolor"] = 'black'
        default_colour_scheme["Task to run"]["fillcolor"] = '"#5f52ee"'
        default_colour_scheme["Task to run"]["fontcolor"] = "lightgrey"
        default_colour_scheme["Up-to-date Final target"]["fontcolor"] = '"#EFA03B"'

    return default_colour_scheme


# _________________________________________________________________________________________
#
#   get_dot_format_for_task_type
# _________________________________________________________________________________________
def get_dot_format_for_task_type(task_type, attributes, colour_scheme, used_formats):
    """
    Look up appropriate colour and style for each type of task
    """
    used_formats.add(task_type)
    for color_type in ("fontcolor", "fillcolor", "color"):
        attributes[color_type] = colour_scheme[task_type][color_type]
        attributes["style"] = "filled"
        if colour_scheme[task_type]["dashed"]:
            attributes["style"] = "dashed"


# _________________________________________________________________________________________
#
#   write_legend_key
# _________________________________________________________________________________________
def write_legend_key(stream, used_task_types, minimal_key_legend, colour_scheme, key_name="Key:", subgraph_index=1):
    """
    Write legend/key to dependency tree graph
    """
    if not len(used_task_types):
        return

    stream.write('subgraph clusterkey%d\n' % subgraph_index)
    stream.write('{\n')

    stream.write('rank="min";\n')
    stream.write('style=filled;\n')
    stream.write('fontsize=20;\n')
    stream.write('color=%s;\n' % (colour_scheme["Key"]["fillcolor"]))
    stream.write('label = "%s";\n' % key_name)
    stream.write('fontcolor = %s;' % (colour_scheme["Key"]["fontcolor"]))
    stream.write('node[margin="0.2,0.2", fontsize="14"];\n')

    #
    #   Only include used task types
    #
    all_task_types = [
        "Vicious cycle",
        "Down stream",
        "Up-to-date task",
        "Explicitly specified task",
        "Task to run",
        "Up-to-date task forced to rerun",
        "Up-to-date Final target",
        "Final target", ]
    if not minimal_key_legend:
        used_task_types |= set(all_task_types)
    wrapped_task_types = [
        "Vicious cycle",
        "Down stream",
        "Up-to-date task",
        "Explicitly specified task",
        "Task to run",
        "Up-to-date task\\nforced to rerun",
        "Up-to-date\\nFinal target",
        "Final target", ]
    wrapped_task_types = dict(zip(all_task_types, wrapped_task_types))

    def outputkey(key, task_type, stream):
        ignore_used_task_types = set()
        attributes = dict()
        attributes["shape"] = "box3d"
        #attributes["shape"] = "rect"
        get_dot_format_for_task_type(
            task_type, attributes, colour_scheme, ignore_used_task_types)
        #attributes["fontsize"] = '15'
        stream.write(key + attributes_to_str(attributes,
                                             wrapped_task_types[task_type]))

    sorted_used_task_types = []
    for t in all_task_types:
        if t in used_task_types:
            sorted_used_task_types.append(t)

    # print first key type
    outputkey("k1_%d" % subgraph_index, sorted_used_task_types[0], stream)

    for i, (from_task_type, to_task_type) in enumerate(adjacent_pairs_iterate(sorted_used_task_types)):
        from_key = 'k%d_%d' % (i + 1, subgraph_index)
        to_key = 'k%d_%d' % (i + 2, subgraph_index)
        # write key
        outputkey(to_key, to_task_type, stream)
        # connection between keys
        stream.write(get_arrow_str_for_legend_key(from_task_type,
                                                  to_task_type, from_key, to_key, colour_scheme))

    stream.write("}\n")


# _________________________________________________________________________________________

#   write_colour_scheme_demo_in_dot_format

# _________________________________________________________________________________________
def write_colour_scheme_demo_in_dot_format(stream):
    """
    Write all the colour schemes in different colours
    """
    stream.write('digraph "Colour schemes"\n{\n')
    stream.write('size="8,11";\n')
    stream.write('splines=true;\n')
    stream.write('fontsize="30";\n')
    stream.write('ranksep = 0.3;\n')
    stream.write('node[fontsize="20"];\n')
    for colour_scheme_index in range(CNT_COLOUR_SCHEMES):
        colour_scheme = get_default_colour_scheme(colour_scheme_index)
        write_legend_key(stream, set("test"), False, colour_scheme,
                         "Colour Scheme %d" % colour_scheme_index, colour_scheme_index)
    stream.write("}\n")


# _________________________________________________________________________________________

#   write_flowchart_in_dot_format

# _________________________________________________________________________________________
def write_flowchart_in_dot_format(jobs_to_run,
                                  up_to_date_jobs,
                                  dag_violating_edges,
                                  dag_violating_nodes,
                                  byte_stream,
                                  target_jobs,
                                  forced_to_run_jobs=[],
                                  all_jobs=None,
                                  vertical=True,
                                  skip_uptodate_tasks=False,
                                  no_key_legend=False,
                                  minimal_key_legend=True,
                                  user_colour_scheme=None,
                                  pipeline_name="Pipeline:"):
    """
    jobs_to_run         = pipeline jobs which are not up to date or have dependencies
                            which are not up to date

    vertical            = print flowchart vertically
    minimal_key_legend  = only print used task types in key legend
    user_colour_scheme  = dictionary for overriding default colours
                          Colours can be names e.g. "black" or quoted hex e.g. '"#F6F4F4"'
                          Default values will be used unless specified

                          key    = "colour_scheme_index": index of default colour scheme
                          key    = "Final target"
                                   "Explicitly specified task"
                                   "Task to run"
                                   "Down stream"
                                   "Up-to-date Final target"
                                   "Up-to-date task forced to rerun"
                                   "Up-to-date task"
                                   "Vicious cycle"
                                   Specifies colours for each task type

                              Subkey = "fillcolor"
                                       "fontcolor"
                                       "color"
                                       "dashed"   = 0/1


                          key    =  "Vicious cycle"
                                    "Task to run"
                                    "Up-to-date"

                              Subkey = "linecolor"
                                       Specifying colours for arrows between tasks

                          key    = "Pipeline"

                              Subkey = "fontcolor"
                                        Specifying flowchart title colour

                          key    = "Key"

                              Subkey = "fontcolor"
                                       "fillcolor"
                                       Specifies legend colours
    """

    if user_colour_scheme is None:
        colour_scheme = get_default_colour_scheme()
    else:
        if "colour_scheme_index" in user_colour_scheme:
            colour_scheme_index = user_colour_scheme["colour_scheme_index"]
        else:
            colour_scheme_index = 0
        colour_scheme = get_default_colour_scheme(colour_scheme_index)
        for k, v in user_colour_scheme.items():
            if k not in colour_scheme:
                continue
            if isinstance(v, dict):
                colour_scheme[k].update(v)
            else:
                colour_scheme[k] = v

    up_to_date_jobs = set(up_to_date_jobs)

    #
    #   cases where child points back to ancestor
    #
    dag_violating_dependencies = set(dag_violating_edges)

    stream = StringIO()

    stream.write('digraph "%s"\n{\n' % pipeline_name)
    stream.write('size="8,11";\n')
    stream.write('splines=true;\n')
    stream.write('fontsize="30";\n')
    stream.write('ranksep = 0.3;\n')
    stream.write('node[fontsize="20"];\n')
    stream.write('graph[clusterrank="local"];\n')

    if not vertical:
        stream.write('rankdir="LR";\n')
    stream.write('subgraph clustertasks\n'
                 "{\n")
    stream.write('rank="min";\n')
    stream.write('fontcolor = %s;\n' % colour_scheme["Pipeline"]["fontcolor"])
    stream.write('label = "%s";\n' % pipeline_name)
    # if vertical:
    #    stream.write( 'edge[minlen=2];\n')
    delayed_task_strings = list()
    vicious_cycle_task_strings = list()

    #
    #   all jobs should be specified
    #       this is a bad fall-back
    #       because there is no guarantee that we are printing what we want to print
    if all_jobs is None:
        all_jobs = node.all_nodes

    used_task_types = set()

    #
    #   defined duplicately in graph. Bad practice
    #
    _one_to_one = 0
    _many_to_many = 1
    _one_to_many = 2
    _many_to_one = 3

    for n in all_jobs:
        attributes = dict()
        attributes["shape"] = "box3d"
        #attributes["shape"] = "rect"
        if hasattr(n, "single_multi_io"):
            if n.single_multi_io == _one_to_many:
                attributes["shape"] = "house"
                attributes["peripheries"] = 2
            elif n.single_multi_io == _many_to_one:
                attributes["shape"] = "invhouse"
                attributes["height"] = 1.1
                attributes["peripheries"] = 2

        #
        #   circularity violating DAG: highlight in red
        #
        if n in dag_violating_nodes:
            get_dot_format_for_task_type(
                "Vicious cycle", attributes, colour_scheme, used_task_types)
            vicious_cycle_task_strings.append(
                't%d' % n._node_index + attributes_to_str(attributes, _get_name(n)))
        #
        #   these jobs will be run
        #
        elif n in jobs_to_run:

            #
            #   up to date but forced to run: outlined in blue
            #
            if n in forced_to_run_jobs:
                get_dot_format_for_task_type(
                    "Explicitly specified task", attributes, colour_scheme, used_task_types)

            #
            #   final target: outlined in orange
            #
            elif n in target_jobs:
                get_dot_format_for_task_type(
                    "Final target", attributes, colour_scheme, used_task_types)

            #
            #   up to date dependency but forced to run: outlined in green
            #
            elif n in up_to_date_jobs:
                get_dot_format_for_task_type(
                    "Up-to-date task forced to rerun", attributes, colour_scheme, used_task_types)

            else:
                get_dot_format_for_task_type(
                    "Task to run", attributes, colour_scheme, used_task_types)

            #
            #   graphviz attributes override other definitions
            #       presume the user knows what she is doing!
            #
            if(hasattr(n, 'graphviz_attributes')):
                for k in n.graphviz_attributes:
                    attributes[k] = n.graphviz_attributes[k]
            stream.write('t%d' % n._node_index +
                         attributes_to_str(attributes, _get_name(n)))

        else:
            #
            #   these jobs are up to date and will not be re-run
            #

            if not skip_uptodate_tasks:
                if n in target_jobs:
                    get_dot_format_for_task_type(
                        "Up-to-date Final target", attributes, colour_scheme, used_task_types)

                elif n in up_to_date_jobs:
                    get_dot_format_for_task_type(
                        "Up-to-date task", attributes, colour_scheme, used_task_types)

                #
                #   these jobs will be ignored: gray with gray dependencies
                #
                else:
                    get_dot_format_for_task_type(
                        "Down stream", attributes, colour_scheme, used_task_types)
                    delayed_task_strings.append(
                        't%d' % n._node_index + attributes_to_str(attributes, _get_name(n)))
                    for o in n._get_inward():
                        delayed_task_strings.append('t%d -> t%d[color=%s, arrowtype=normal];\n' % (
                            o._node_index, n._node_index, colour_scheme["Up-to-date"]["linecolor"]))
                    continue

            #
            #   graphviz attributes override other definitions
            #       presume the user knows what she is doing!
            #
            if(hasattr(n, 'graphviz_attributes')):
                for k in n.graphviz_attributes:
                    attributes[k] = n.graphviz_attributes[k]
            stream.write('t%d' % n._node_index +
                         attributes_to_str(attributes, _get_name(n)))

        #
        #   write edges
        #
        unconstrained = False
        for o in sorted(n._get_inward(), reverse=True, key=lambda x: x._node_index):
            #
            #   circularity violating DAG: highlight in red: should never be a constraint
            #       in drawing the graph
            #
            if (n, o) in dag_violating_dependencies:
                constraint_str = ", constraint=false" if o._node_index > n._node_index else ""
                vicious_cycle_task_strings.append('t%d -> t%d[color=%s %s];\n' % (
                    o._node_index, n._node_index, colour_scheme["Vicious cycle"]["linecolor"], constraint_str))
                continue
            elif not o in jobs_to_run or not n in jobs_to_run:
                if not skip_uptodate_tasks:
                    edge_str = 't%d -> t%d[color=%s, arrowtype=normal];\n' % (
                        o._node_index, n._node_index, colour_scheme["Up-to-date"]["linecolor"])
                    if unconstrained:
                        delayed_task_strings.append(edge_str)
                    else:
                        stream.write(edge_str)
            else:
                stream.write('t%d -> t%d[color=%s];\n' % (o._node_index,
                                                          n._node_index, colour_scheme["Task to run"]["linecolor"]))
            unconstrained = True

    for l in delayed_task_strings:
        stream.write(l)

    #
    #   write vicious cycle at end so not constraint in drawing graph
    #
    for l in vicious_cycle_task_strings:
        stream.write(l)
    stream.write('}\n')

    if not no_key_legend:
        write_legend_key(stream, used_task_types,
                         minimal_key_legend, colour_scheme)
    stream.write("}\n")

    ss = stream.getvalue().encode()
    byte_stream.write(ss)

    stream.close()



from __future__ import print_function
################################################################################
#
#   graph.py
#
#
#   Copyright (c) 10/9/2009 Leo Goodstadt
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#################################################################################
"""
    graph.py

        provides support for diacyclic graph
        with topological_sort

"""
import sys
import re
import os

# use simplejson in place of json for python < 2.6
try:
    import json
except ImportError:
    import simplejson
    json = simplejson

from collections import defaultdict
from .print_dependencies import *
import tempfile
import subprocess
# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   class node

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


class graph_error(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class error_duplicate_node_name(graph_error):
    pass


class node (object):
    """
    node
        designed for diacyclic graphs but can hold anything
        contains lists of nodes and
        dictionary to look up node name from node
    """

    _all_nodes = list()
    _name_to_node = dict()
    _index_to_node = dict()
    _global_node_index = 0

    _one_to_one = 0
    _many_to_many = 1
    _one_to_many = 2
    _many_to_one = 3

    @staticmethod
    def _get_leaves():
        for n in node._all_nodes:
            if len(n._inward) == 0:
                yield n

    @staticmethod
    def _get_roots():
        for n in node._all_nodes:
            if len(n._outward) == 0:
                yield n

    @staticmethod
    def _count_nodes():
        return len(_all_nodes)

    @staticmethod
    def _dump_tree_as_str():
        """
        dumps entire tree
        """
        return ("%d nodes " % node._count_nodes()) + "\n" + \
            "\n".join([x._fullstr() for x in node._all_nodes])

    @staticmethod
    def _lookup_node_from_name(name):
        return node._name_to_node[name]

    @staticmethod
    def _lookup_node_from_index(index):
        return node._index_to_node[index]

    @staticmethod
    def _is_node(name):
        return name in node._name_to_node

    # _____________________________________________________________________________________

    #   init

    # _____________________________________________________________________________________

    def __init__(self, name, **args):
        """
        each node has
            _name
            _inward :   lists of incoming edges
            _outward:   lists of outgoing edges
        """
        #
        #   make sure node name is unique
        #
        # if name in node._name_to_node:
        #    raise error_duplicate_node_name("[%s] has already been added" % name)

        self.__dict__.update(args)
        self._inward = list()
        self._outward = list()
        self.args = args
        self._name = name
        self._signal = False
        self._node_index = node._global_node_index
        node._global_node_index += 1

        #
        #   for looking up node for name
        #
        node._all_nodes.append(self)
        node._name_to_node[self._name] = self
        node._index_to_node[self._node_index] = self

    # _____________________________________________________________________________________

    #   _add_child

    # _____________________________________________________________________________________
    def _add_child(self, child):
        """
        connect edges
        """
        # do not add duplicates
        if child in self._outward:
            return child

        self._outward.append(child)
        child._inward.append(self)
        return child

    # _____________________________________________________________________________________

    #   _remove_child

    # _____________________________________________________________________________________
    def _remove_child(self, child):
        """
        disconnect edges
        """

        if child in self._outward:
            self._outward.remove(child)
        if self in child._inward:
            child._inward.remove(self)
        return child
    # _____________________________________________________________________________________

    #   _add_parent

    # _____________________________________________________________________________________
    def _add_parent(self, parent):
        """
        connect edges
        """
        # do not add duplicates
        if parent in self._inward:
            return parent

        self._inward.append(parent)
        parent._outward.append(self)
        return parent

    # _____________________________________________________________________________________

    #   _remove_all_parents

    # _____________________________________________________________________________________
    def _remove_all_parents(self):
        """
        disconnect edges
        """

        # remove self from parent
        for parent in self._inward:
            if self in parent._outward:
                parent._outward.remove(self)

        # clear self
        self._inward = []

        return self

    # _____________________________________________________________________________________

    #   _remove_parent

    # _____________________________________________________________________________________
    def _remove_parent(self, parent):
        """
        disconnect edges
        """

        if parent in self._inward:
            self._inward.remove(parent)
        if self in parent._outward:
            parent._outward.remove(self)
        return parent
    # _____________________________________________________________________________________

    #   _get_inward/_get_outward

    # _____________________________________________________________________________________
    def _get_outward(self):
        """
        just in case we need to return inward when we mean outward!
            (for reversed graphs)
        """
        return self._outward

    def _get_inward(self):
        """
        just in case we need to return inward when we mean outward!
            (for reversed graphs)
        """
        return self._inward

    # _____________________________________________________________________________________

    #   _fullstr

    # _____________________________________________________________________________________

    def _fullstr(self):
        """
        Full dump. Normally edges are not printed out
        Everything is indented except name
        """
        self_desc = list()
        for k, v in sorted(iter(self.__dict__.items()), key=lambda x_v: (0, x_v[0], x_v[1]) if x_v[0] == "_name" else (1, x_v[0], x_v[1])):
            indent = "    " if k != "_name" else ""
            if k in ("_inward", "_outward"):
                v = ",".join([x._name for x in v])
                self_desc.append(indent + str(k) + "=" + str(v))
            else:
                self_desc.append(indent + str(k) + "=" + str(v))
        return "\n".join(self_desc)

    # _____________________________________________________________________________________

    #   __str__

    # _____________________________________________________________________________________
    def __str__(self):
        """
        Print everything except lists of edges
        Useful for debugging
        """
        self_desc = list()
        for k, v in sorted(self.__dict__.items(), reverse=True):
            indent = "    " if k != "_name" else ""
            if k[0] == '_':
                continue
            else:
                self_desc.append(indent + str(k) + "=" + str(v))
        return "  Task = " + "\n".join(self_desc)

    # _____________________________________________________________________________________

    #   _signalled
    #
    # _____________________________________________________________________________________

    def _signalled(self, extra_data_for_signal=None):
        """
        Signals whether depth first search ends without this node
        """
        return self._signal


# _____________________________________________________________________________________

#   node_to_json
#
#
# _____________________________________________________________________________________
class node_to_json(json.JSONEncoder):
    """
    output node using json
    """

    def default(self, obj):
        print(str(obj))
        if isinstance(obj, node):
            return obj._name, {
                "index": obj._node_index,
                "_signal": obj._signal,
                "_get_inward": [n._name for n in obj._inward],
                "_get_outward": [n._name for n in obj._outward],
            }
        return json.JSONEncoder.default(self, obj)


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   topological_sort_visitor

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


def default_signalled(node, extra_data):
    """
    Depth first search stops when node._signalled return True
    """
    return node._signalled(extra_data)


class topological_sort_visitor (object):
    """
    topological sort
        used with DFS to find all nodes in topologically sorted order
        All finds all DAG breaking cycles

    """

    IGNORE_NODE_SIGNAL = 0
    NOTE_NODE_SIGNAL = 1
    END_ON_SIGNAL = 2

    # _____________________________________________________________________________________

    #   init

    # _____________________________________________________________________________________
    def __init__(self, forced_dfs_nodes,
                 node_termination=END_ON_SIGNAL,
                 extra_data_for_signal=None,
                 signal_callback=None):
        """
        list of saved results
        """
        self._forced_dfs_nodes = set(forced_dfs_nodes)
        self._node_termination = node_termination

        self._start_nodes = set()
        self._back_edges = set()
        self._back_nodes = set()
        self._signalling_nodes = set()

        self.signal_callback = signal_callback

        # keep order for tree traversal later
        self._examined_edges = list()
        # keep order for topological sorted results
        self._finished_nodes = list()

        self._extra_data_for_signal = extra_data_for_signal

    def combine_with(self, other):
        """
        combine the results of two visitors
            (add other to self)
        """
        self._back_edges              .update(other._back_edges)
        self._back_nodes              .update(other._back_nodes)
        extra_finished_nodes = set(
            other._finished_nodes) - set(self._finished_nodes)
        self._finished_nodes          .extend(extra_finished_nodes)

    # _____________________________________________________________________________________

    #   __str__

    # _____________________________________________________________________________________

    def __str__(self):
        """
        for diagnostics
        """
        signalling_str = get_nodes_str("Signalling", self._signalling_nodes)
        finished_str = get_nodes_str("Finished", self._finished_nodes)
        forced_str = get_nodes_str("Forced to run", self._forced_dfs_nodes)
        start_str = get_nodes_str("Start", self._start_nodes)
        back_edges_str = get_edges_str("back", self._back_edges)
        return (""
                + finished_str
                + start_str
                + back_edges_str
                + signalling_str
                + finished_str
                )

    # _____________________________________________________________________________________

    #   not_dag

    # _____________________________________________________________________________________
    def not_dag(self):
        """
        back edges add circularity
        """
        return len(self._back_edges)

    # _____________________________________________________________________________________

    #   dag_violating_edges

    # _____________________________________________________________________________________
    def dag_violating_edges(self):
        """
        back edges add circularity
        """
        return self._back_edges

    # _____________________________________________________________________________________

    #   dag_violating_nodes

    # _____________________________________________________________________________________

    def dag_violating_nodes(self):
        """
        all nodes involved in cycless
        """
        return self._back_nodes

    # _____________________________________________________________________________________

    #   identify_dag_violating_nodes_and_edges
    #
    # _____________________________________________________________________________________
    def identify_dag_violating_nodes_and_edges(self):
        """
        find all nodes and edges in any cycles

        All dag violating cycles are defined by the back edge identified in DFS.
        All paths which go the other way: start at the to_node and end up at the from_node
            are therefore also part of the cycle

        """
        if not len(self._back_edges):
            return
        cnt_examined_edges = len(self._examined_edges)

        # add this to _back_edges at the end
        cycle_edges = set()

        #
        #   each cycle
        #       starts from the to_node   of each back_edge and
        #       ends   with the from_node of each back_edge
        #
        for cycle_to_node, cycle_from_node in self._back_edges:
            start_search_from = 0
            while 1:
                #
                # find start of cycle
                for i, (f, t, n) in enumerate(self._examined_edges[start_search_from:]):
                    if f == cycle_from_node:
                        break

                # no more cycles for this cycle_from_node/cycle_to_node pair
                else:
                    break

                #
                # cycle end might be within the same pair
                #   if so, don't search the current (not the next) edge for the cycle end
                #
                #   Otherwise incrementing search position avoids infinite loop
                #
                start_search_from = cycle_start = start_search_from + i
                if self._examined_edges[cycle_start][1] != cycle_to_node:
                    start_search_from += 1

                for i, (f, t, n) in enumerate(self._examined_edges[start_search_from:]):

                    #
                    #   found end of cycle
                    #
                    if t == cycle_to_node:
                        cycle_end = start_search_from + i + 1
                        #
                        #   ignore backtracked nodes which will not be part of the cycle
                        #       we are essentially doing tree traversal here
                        #
                        backtracked_nodes = set()
                        for f, t, n in self._examined_edges[cycle_start:cycle_end]:
                            if t is None:
                                backtracked_nodes.add(n)
                        for f, t, n in self._examined_edges[cycle_start:cycle_end]:
                            if f is None or f in backtracked_nodes or t in backtracked_nodes:
                                continue
                            cycle_edges.add((f, t))
                            self._back_nodes.add(f)
                            self._back_nodes.add(t)
                        start_search_from = cycle_end
                        break

                    # if cycle_from_node comes around again, this is not a cycle
                    if cycle_from_node == f:
                        if not i:
                            i += 1
                        start_search_from = start_search_from + i
                        break

                    continue

                # no more cycles for this cycle_from_node/cycle_to_node pair
                else:
                    break

        self._back_edges.update(cycle_edges)

    # _____________________________________________________________________________________

    #   not_dag

    # _____________________________________________________________________________________

    def topological_sorted(self):
        """
        _finished_nodes
        """
        return self._finished_nodes

    # _____________________________________________________________________________________

    #   terminate_before

    # _____________________________________________________________________________________

    def terminate_before(self, node):
        """
        Allow node to terminate this path in DFS without including itself
            (see terminate_at)

        If node in _forced_dfs_nodes that overrides what the node wants
        """

        #
        #   If _node_termination = IGNORE_NODE_TERMINATION
        #       always go through whole tree
        #
        if self._node_termination == self.IGNORE_NODE_SIGNAL:
            return False

        #
        #   If _node_termination = NOTE_NODE_TERMINATION
        #       always go through whole tree but remember
        #       which nodes want to terminate
        #
        #   Note that _forced_dfs_nodes is ignored
        #
        if self._node_termination == self.NOTE_NODE_SIGNAL:
            if self.signal_callback(node, self._extra_data_for_signal):
                self._signalling_nodes.add(node)
            return False

        #
        #   _forced_dfs_nodes always overrides node preferences
        #       but let us save what the node says anyway for posterity
        #
        if node in self._forced_dfs_nodes:
            # Commented out code lets us save self_terminating_nodes even when
            # they have been overridden by _forced_dfs_nodes
            # if self.signal_callback(node, self._extra_data_for_signal):
            #    self._signalling_nodes.add(node)
            return False

        #
        #   OK. Go by what the node wants then
        #
        if self.signal_callback(node, self._extra_data_for_signal):
            self._signalling_nodes.add(node)
            return True
        return False

    # _____________________________________________________________________________________

    #   call_backs

    # _____________________________________________________________________________________

    def discover_vertex(self, node):
        pass

    def start_vertex(self, node):
        self._start_nodes.add(node)

    def finish_vertex(self, node):
        """
        Save
            1) topologically sorted nodes
            2) as "None" (back) edges which allows _examined_edges to be traversed
               like a tree

        """
        self._examined_edges.append((None, None, node))
        self._finished_nodes.append(node)

    def examine_edge(self, node_from, node_to):
        """
        Save edges as we encounter then so we can look for loops

        """
        self._examined_edges.append((node_from, node_to, None))

    def back_edge(self, node_from, node_to):
        self._back_edges.add((node_from, node_to))

    def tree_edge(self, node_from, node_to):
        pass

    def forward_or_cross_edge(self, node_from, node_to):
        pass

    def terminate_at(self, node):
        """
        Terminate this line of DFS but include myself
        """
        return False

#
# _________________________________________________________________________________________

#   debug_print_visitor

# _________________________________________________________________________________________


class debug_print_visitor (object):
    """
    log progress through DFS: for debugging

    """

    def terminate_before(self, node):
        return False

    def terminate_at(self, node):
        return False

    def start_vertex(self, node):
        print("s  start vertex %s" % (node._name))

    def finish_vertex(self, node):
        print("  v  finish vertex %s" % (node._name))

    def discover_vertex(self, node):
        print("  |  discover vertex %s" % (node._name))

    def examine_edge(self, node_from, node_to):
        print("  -- examine edge %s -> %s" % (node_from._name, node_to._name))

    def back_edge(self, node_from, node_to):
        print("    back edge %s -> %s" % (node_from._name, node_to._name))

    def tree_edge(self, node_from, node_to):
        print("   - tree edge %s -> %s" % (node_from._name, node_to._name))

    def forward_or_cross_edge(self, node_from, node_to):
        print("   - forward/cross edge %s -> %s" %
              (node_from._name, node_to._name))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#   Functions
# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
# _________________________________________________________________________________________
#   depth first search
# _________________________________________________________________________________________
#
#
#
WHITE = 0       # virgin
GRAY = 1       # processing
BLACK = 2       # finished


def depth_first_visit(u, visitor, colours, outedges_func):
    """
    depth_first_visit
        unused callbacks are commented out
    """
    # start processing this node: so gray
    colours[u] = GRAY

    stack = list()

    #
    # unused callback
    #
    # visitor.discover_vertex(u)

    curr_edges = outedges_func(u)

    if visitor.terminate_before(u):
        colours[u] = BLACK
        return
    # If this vertex terminates the search, we push empty range
    if visitor.terminate_at(u):
        stack.append((u, curr_edges, len(curr_edges)))
    else:
        stack.append((u, curr_edges, 0))

    while len(stack):
        u, curr_edges, curr_edge_pos = stack.pop()
        while curr_edge_pos < len(curr_edges):
            v = curr_edges[curr_edge_pos]
            visitor.examine_edge(u, v)
            v_colour = colours[v]

            if visitor.terminate_before(v):
                colours[v] = BLACK
                curr_edge_pos += 1
                continue

            if v_colour == WHITE:
                #
                # unused callback
                #
                #visitor.tree_edge(u, v)
                curr_edge_pos += 1
                stack.append((u, curr_edges, curr_edge_pos))
                u = v
                colours[u] = GRAY
                #
                # unused callback
                #
                # visitor.discover_vertex(u)
                curr_edges = outedges_func(u)
                curr_edge_pos = 0

                if visitor.terminate_at(u):
                    break
            elif v_colour == GRAY:
                visitor.back_edge(u, v)
                curr_edge_pos += 1
            else:
                #
                # unused callback
                #
                #visitor.forward_or_cross_edge(u, v)
                curr_edge_pos += 1
        colours[u] = BLACK
        visitor.finish_vertex(u)


def depth_first_search(starting_nodes, visitor, outedges_func=node._get_inward):
    """
    depth_first_search
        go through all starting points and DFV on each of them
        if they haven't been seen before
    """
    colours = defaultdict(int)  # defaults to WHITE
    if len(starting_nodes):
        for start in starting_nodes:
            if colours[start] == WHITE:
                visitor.start_vertex(start)
                depth_first_visit(start, visitor, colours, outedges_func)
    else:

        #
        #   go through all nodes, maintaining order
        #
        for start in node._all_nodes:
            if colours[start] == WHITE:
                visitor.start_vertex(start)
                depth_first_visit(start, visitor, colours, outedges_func)


def topologically_sorted_nodes(to_leaves,
                               force_start_from=[],
                               gather_all_non_signalled=True,
                               test_all_signals=False,
                               extra_data_for_signal=None,
                               signal_callback=None):
    """
    Get all nodes which are children of to_leaves
        in topological sorted order

    Defaults to including all nodes which are non-signalled and their dependents (via include_any_children())
        i.e. includes the *last* non-signalling node on each branch and all the way up the tree


    Otherwise stops at each branch just before signalling node
        i.e. includes the last non-signalling *run* on each branch


    force_start_from
    Optionally specify all the child nodes which *have* to
        be included in the list at least
        This will override any node signals

    force_start_from = True to get the whole tree irrespective of signalling


    Rewritten to minimise calls to node._signalled()

    """

    #
    #   got through entire tree, looking for signalling nodes,
    #       usually for debugging or for printing
    #
    if test_all_signals:
        v = topological_sort_visitor([],
                                     topological_sort_visitor.NOTE_NODE_SIGNAL,
                                     extra_data_for_signal,
                                     signal_callback)
        depth_first_search(to_leaves, v, node._get_inward)
        signalling_nodes = v._signalling_nodes
    else:
        signalling_nodes = set()

    if gather_all_non_signalled:
        #
        #   get whole tree, ignoring signalling
        #
        v = topological_sort_visitor(
            [], topological_sort_visitor.IGNORE_NODE_SIGNAL)
        depth_first_search(to_leaves, v, node._get_inward)

        #
        #   not dag: no further processing
        #
        if v.not_dag():
            v.identify_dag_violating_nodes_and_edges()
            return (v.topological_sorted(), v._signalling_nodes, v.dag_violating_edges(),
                    v.dag_violating_nodes())

        #
        #   if force_start_from == True
        #
        #       return entire tree
        #
        if force_start_from == True:
            return (v.topological_sorted(), v._signalling_nodes, v.dag_violating_edges(),
                    v.dag_violating_nodes())

        #
        #   Set of all nodes we are going to return
        #   We will use v.topological_sorted to return them in the right (sorted) order
        #
        nodes_to_include = set()

        #
        #   If force start from is a list of nodes,
        #       include these and all of its dependents (via include_any_children)
        #
        #   We don't need to bother to check if they signal (_signalled)
        #   This saves calling the expensive _signalled
        #
        if len(force_start_from):
            nodes_to_include.update(include_any_children(force_start_from))
        # This should not be necessary because include_any_children also returns self.
        # for n in force_start_from:
        #    if n in nodes_to_include:
        #        continue
        #    nodes_to_include.add(n)
        #    nodes_to_include.update(include_any_children([n]))

        #
        #   Now select all nodes from ancestor -> descendant which do not signal (signal_callback() == false)
        #       and select their descendants (via include_any_children())
        #
        #   Nodes which signal are added to signalling_nodes
        #
        reversed_nodes = v.topological_sorted()
        for n in reversed_nodes:
            if n in nodes_to_include:
                continue

            if not signal_callback(n, extra_data_for_signal):
                # nodes_to_include.add(n)
                nodes_to_include.update(include_any_children([n]))
            else:
                signalling_nodes.add(n)
                #sys.stderr.write(json.dumps(n, cls=node_to_json, sort_keys=1) + "\n")

        return ([n for n in v.topological_sorted() if n in nodes_to_include],
                signalling_nodes,
                [], [])

    #
    #   gather_all_non_signalled = False
    #       stop at first signalled
    #
    else:

        if force_start_from == True:
            #
            #   get whole tree, ignoring signalling
            #
            v = topological_sort_visitor([],
                                         topological_sort_visitor.IGNORE_NODE_SIGNAL)
        else:
            #
            #   End at each branch without including signalling node
            #       but ignore signalling for forced_nodes_and_dependencies
            #

            #   Get forced nodes and all descendants via include_any_children
            #
            forced_nodes_and_dependencies = []
            if len(force_start_from):
                forced_nodes_and_dependencies = include_any_children(
                    force_start_from)

            v = topological_sort_visitor(forced_nodes_and_dependencies,
                                         topological_sort_visitor.END_ON_SIGNAL,
                                         extra_data_for_signal,
                                         signal_callback)

        #
        #   Forward graph iteration
        #
        depth_first_search(to_leaves, v, node._get_inward)

        if v.not_dag():
            v.identify_dag_violating_nodes_and_edges()

        signalling_nodes.update(v._signalling_nodes)
        return (v.topological_sorted(), signalling_nodes, v.dag_violating_edges(), v.dag_violating_nodes())


#
def debug_print_nodes(to_leaves):
    v = debug_print_visitor()
    depth_first_search(to_leaves, v, node._get_inward)


# _________________________________________________________________________________________

#   graph_printout

# _________________________________________________________________________________________
def graph_colour_demo_printout(stream,
                               output_format,
                               size='11,8',
                               dpi='120'):
    """
    Demo of the different colour schemes
    """

    if output_format == 'dot':
        write_colour_scheme_demo_in_dot_format(stream)
        return

    # print to dot file
    #temp_dot_file = tempfile.NamedTemporaryFile(suffix='.dot', delete=False)
    fh, temp_dot_file_name = tempfile.mkstemp(suffix='.dot')
    temp_dot_file = os.fdopen(fh, "w")

    write_colour_scheme_demo_in_dot_format(temp_dot_file)
    temp_dot_file.close()

    print_dpi = ("-Gdpi='%s'" % dpi) if output_format != "svg" else ""
    run_dot = os.popen("dot -Gsize='%s' %s -T%s < %s" %
                       (size, print_dpi, output_format, temp_dot_file_name))

    #
    #   wierd bug fix for firefox and svg
    #
    result_str = run_dot.read()
    err = run_dot.close()
    if err:
        raise RuntimeError("dot failed to run with exit code %d" % err)
    if output_format == "svg":
        result_str = result_str.replace("0.12", "0.0px")
    stream.write(result_str)
# _________________________________________________________________________________________

#   graph_printout_in_dot_format

# _________________________________________________________________________________________


def graph_printout_in_dot_format(stream,
                                 to_leaves,
                                 force_start_from=[],
                                 draw_vertically=True,
                                 ignore_upstream_of_target=False,
                                 skip_signalling_nodes=False,
                                 gather_all_non_signalled=True,
                                 test_all_signals=True,
                                 no_key_legend=False,
                                 minimal_key_legend=True,
                                 user_colour_scheme=None,
                                 pipeline_name="Pipeline:",
                                 extra_data_for_signal=None,
                                 signal_callback=None):
    """
    print out pipeline dependencies in dot formatting
    """

    (topological_sorted,        # tasks_to_run
     signalling_nodes,           # up to date
     dag_violating_edges,
     dag_violating_nodes) = topologically_sorted_nodes(to_leaves, force_start_from,
                                                       gather_all_non_signalled,
                                                       test_all_signals,
                                                       extra_data_for_signal,
                                                       signal_callback)

    #
    #   N.B. For graph:
    #           upstream   = parent
    #           dependents/downstream
    #                      = children
    #
    #
    nodes_to_display = get_reachable_nodes(
        to_leaves, not ignore_upstream_of_target)

    #
    #   print out dependencies in dot format
    #
    write_flowchart_in_dot_format(topological_sorted,           # tasks_to_run
                                  signalling_nodes,             # up to date
                                  dag_violating_edges,
                                  dag_violating_nodes,
                                  stream,
                                  to_leaves,
                                  force_start_from,
                                  nodes_to_display,
                                  draw_vertically,
                                  skip_signalling_nodes,
                                  no_key_legend,
                                  minimal_key_legend,
                                  user_colour_scheme,
                                  pipeline_name)

# _________________________________________________________________________________________

#   graph_printout

# _________________________________________________________________________________________


def graph_printout(stream,
                   output_format,
                   to_leaves,
                   force_start_from=[],
                   draw_vertically=True,
                   ignore_upstream_of_target=False,
                   skip_signalling_nodes=False,
                   gather_all_non_signalled=True,
                   test_all_signals=True,
                   no_key_legend=False,
                   minimal_key_legend=True,
                   user_colour_scheme=None,
                   pipeline_name="Pipeline:",
                   size=(11, 8),
                   dpi=120,
                   extra_data_for_signal=None,
                   signal_callback=None):
    """
    print out pipeline dependencies in a variety of formats, using the programme "dot"
        an intermediary
    """

    if output_format == 'dot':
        graph_printout_in_dot_format(stream,
                                     to_leaves,
                                     force_start_from,
                                     draw_vertically,
                                     ignore_upstream_of_target,
                                     skip_signalling_nodes,
                                     gather_all_non_signalled,
                                     test_all_signals,
                                     no_key_legend,
                                     minimal_key_legend,
                                     user_colour_scheme,
                                     pipeline_name,
                                     extra_data_for_signal,
                                     signal_callback)
        return

    # print to dot file
    #temp_dot_file = tempfile.NamedTemporaryFile(suffix='.dot', delete=False)
    fh, temp_dot_file_name = tempfile.mkstemp(suffix='.dot')
    temp_dot_file = os.fdopen(fh, "wb")

    graph_printout_in_dot_format(temp_dot_file,
                                 to_leaves,
                                 force_start_from,
                                 draw_vertically,
                                 ignore_upstream_of_target,
                                 skip_signalling_nodes,
                                 gather_all_non_signalled,
                                 test_all_signals,
                                 no_key_legend,
                                 minimal_key_legend,
                                 user_colour_scheme,
                                 pipeline_name,
                                 extra_data_for_signal,
                                 signal_callback)
    temp_dot_file.close()

    if isinstance(size, tuple):
        print_size = "(%d,%d)" % size
    elif isinstance(size, (str)):
        print_size = size
    else:
        raise Exception(
            "Flowchart print size [%s] should be specified as a tuple of X,Y in inches" % str(size))

    #
    #   N.B. Resolution doesn't seem to play nice with SVG and is ignored
    #
    print_dpi = ("-Gdpi='%s'" % dpi) if output_format != "svg" else ""
    cmd = "dot -Gsize='%s' %s -T%s < %s" % (print_size,
                                            print_dpi, output_format, temp_dot_file_name)

    proc = subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    result_str, error_str = proc.communicate()
    retcode = proc.returncode
    if retcode:
        raise subprocess.CalledProcessError(
            retcode, cmd + "\n" + "\n".join([str(result_str), str(error_str)]))

    #run_dot = os.popen(cmd)
    #result_str = run_dot.read()
    #err = run_dot.close()
    #
    # if err:
    #    raise RuntimeError("dot failed to run with exit code %d" % err)

    #
    #   wierd workaround for bug / bad interaction between firefox and svg:
    #           Font sizes have "px" appended.
    #

    if output_format == "svg":
        # result str is a binary string. I.e. could be .jpg
        # must turn it into string before we can replace, and then turn it back into binary
        result_str = result_str.decode()
        result_str = result_str.replace("0.12", "0.0px")
        result_str = result_str.encode()
    stream.write(result_str)


# _________________________________________________________________________________________

#   include_any_children

# _________________________________________________________________________________________
def include_any_children(nodes):
    """
    Get all children nodes by DFS in the inward direction,
    Ignores signals
    Also includes original nodes in the results
    """
    children_visitor = topological_sort_visitor(
        [], topological_sort_visitor.IGNORE_NODE_SIGNAL)
    depth_first_search(nodes, children_visitor, node._get_outward)
    return children_visitor.topological_sorted()


# _________________________________________________________________________________________

#   get_reachable_nodes

# _________________________________________________________________________________________
def get_reachable_nodes(nodes, children_as_well=True):
    """
    Get all nodes which are parents and children of nodes
        recursing through the entire tree

        i.e. go up *and* down tree starting from node

    1) specify parents_as_well = False
        to only get children and not parents of nodes
        """

    # look for parents of nodes and start there instead
    if children_as_well:
        nodes = include_any_children(nodes)

    parent_visitor = topological_sort_visitor(
        [], topological_sort_visitor.IGNORE_NODE_SIGNAL)
    depth_first_search(nodes, parent_visitor, node._get_inward)
    return parent_visitor.topological_sorted()


# _________________________________________________________________________________________

#   Helper functions to dump edges and nodes

# _________________________________________________________________________________________
def get_edges_str(name, edges):
    """
    helper function to dump edges as a list of names
    """
    edges_str = "  %d %s edges\n" % (len(edges), name)
    edges_str += "    " + \
        ", ".join([x_y[0]._name + "->" + x_y[1]._name for x_y in edges]) + "\n"
    return edges_str


def get_nodes_str(name, nodes):
    """
    helper function to dump nodes as a list of names
    """
    nodes_str = "  %s nodes = %d\n" % (name, len(nodes))
    nodes_str += "    " + ", ".join([x._name for x in nodes]) + "\n"
    return nodes_str



#!/usr/bin/env python
from __future__ import print_function
import sys
if sys.hexversion < 0x03000000:
    from future_builtins import zip
################################################################################
#
#   adjacent_pairs_iterate.py
#
#
#   Copyright (c) 2007 Leo Goodstadt
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#################################################################################


def adjacent_pairs_iterate(array, reverse=False):
    """
    returns pairs of iterators to successive positions
    """
    if len(array) < 2:
        return

    # get iterators to successive positions
    if reverse:
        curr_iter = reversed(array)
        next_iter = reversed(array)
    else:
        curr_iter = iter(array)
        next_iter = iter(array)
    next(next_iter)
    for i, j in zip(curr_iter, next_iter):
        yield i, j


def unit_test():
    numbers = list(range(10))
    print("Forward")
    for i, j in adjacent_pairs_iterate(numbers):
        print(i, j)
    print("Reversed")
    for i, j in adjacent_pairs_iterate(numbers, reverse=True):
        print(i, j)
    print(numbers)


if __name__ == '__main__':
    unit_test()
